//
// $Id: Brain.h,v 1.2 1998/09/03 02:04:03 nygard Exp $
//

#import <AppKit/AppKit.h>

@class FIDock;
@class FIDockView;
@class FITile;

@class TileSettingsController;

@interface Brain : NSObject
{
    IBOutlet NSMenuItem *levelsMenuItem;

    FIDock *dock;

    // The Brain will act as the controller and app delegate.
    NSWindow *dockWindow;
    FIDockView *dockView;

    // Making these singletons would mean we don't need the ivars here.
    TileSettingsController *tileSettingsController;

    NSTimer *saveTimer;

    NSMutableArray *selectedTiles;
}

+ (void) initialize;

- init;
- (void) dealloc;

- (void) applicationDidFinishLaunching:(NSNotification *)aNotification;
- (void) applicationWillTerminate:(NSNotification *)aNotification;

- (void) workspaceNotification:(NSNotification *)aNotification;

- (void) showInfoPanel:sender;
- (void) showPreferencePanel:sender;
- (void) showActiveApplications:sender;

- (void) goToLevel:sender;
//- (void) findLevel:sender;
//- (void) findIcon:sender;

- (void) addWorkspaceIcon:sender;
- (void) placeLevelFirst:sender;
- (void) placeLevelLast:sender;
//- (void) moveLevel:sender;
- (void) rotateLevel:sender;
- (void) nameLevel:sender;
- (void) addLevel:sender;
- (void) clearLevel:sender;
- (void) deleteLevel:sender;
//- (void) moveIcons:sender;

- (void) inspectSelectedIcon:sender;

- (void) showReleaseNotes:sender;
- (void) showInstructions:sender;


- (void) go:sender;

//----------------------------------------------------------------------
// NSApplication delegate
//----------------------------------------------------------------------

- (void) applicationDidBecomeActive:(NSNotification *)aNotification;
- (void) applicationWillBecomeActive:(NSNotification *)aNotification;

//----------------------------------------------------------------------
// Document handling
//----------------------------------------------------------------------

- (NSString *) screenSizeString;
- (NSString *) dockFilename;
- (NSString *) dockPath;

- (void) loadDock;
- (void) saveDockIfNeeded;
- (void) saveDock;

// other?

- (void) createDockWindow;
- (void) showDock;

//----------------------------------------------------------------------
// Save timer
//----------------------------------------------------------------------

- (void) startSaveTimer;
- (void) stopSaveTimer;
- (void) fireSaveTimer:(NSTimer *)timer;

//----------------------------------------------------------------------
// FIDock delegate
//----------------------------------------------------------------------

- (void) updateChangeCount;
- (void) updateDockView;

//----------------------------------------------------------------------
// FIDockView Delegate Protocol
//----------------------------------------------------------------------

- (void) nextLevel:sender;
- (void) previousLevel:sender;
- (void) activateApplication:sender;
- (void) addApplicationsToCurrentLevel:(NSArray *)applications;
- (void) toggleDockVisible:sender;
- (void) toggleDockSunken:sender;
- (void) toggleDockLocked:sender;

- (void) dockWindowDidDrag:(NSPoint)newWindowOrigin;

//----------------------------------------------------------------------
// NSWindow delegate (Dock Window)
//----------------------------------------------------------------------

- (void) windowDidMove:(NSNotification *)aNotification;

//----------------------------------------------------------------------
// Selection
//----------------------------------------------------------------------

- (void) selectTile:(FITile *)tile;
- (void) deselectTile:(FITile *)tile;
- (void) toggleSelection:sender;

- (void) deleteSelectedTiles;

- (void) selectAll:sender;
- (void) deselectAllTiles:sender;

//----------------------------------------------------------------------
// Cut & Paste
//----------------------------------------------------------------------

- (void) cut:sender;
- (void) copy:sender;
- (void) paste:sender;

//----------------------------------------------------------------------
// Other
//----------------------------------------------------------------------

- (NSPoint) initialFiendTilePoint;

- (BOOL) validateMenuItem:(NSMenuItem *)menuCell;

- (void) rebuildLevelsMenu;
- (void) dockLevelsDidChange:(NSNotification *)aNotification;

//----------------------------------------------------------------------
// Services
//----------------------------------------------------------------------

- (void) submitToFiendDock:(NSPasteboard *)pasteboard userData:(NSData *)data error:(NSString **)error;

@end
