#import "Fiend.h"

RCSID ("$Id: Brain.m,v 1.2 1998/09/03 02:04:05 nygard Exp $");

#import "Brain.h"

#import "FIDock.h"
#import "NameLevelController.h"
#import "ActiveApplicationsController.h"
#import "TileSettingsController.h"
#import "NSNotificationCenter+FIExtensions.h"
#import "PreferencePanelController.h"
#import "InfoPanelController.h"

#import "Fiendish.h"
#import "FIAppInfo.h"
#import "FIDockView.h"
#import "FIWindow.h"
#import "FITile.h"
#import "FILevel.h"
#import "FISoundManager.h"

#import "FiendWraps.h"

@interface NSWindow (hidden)
- (void) _setAvoidsActivation:(BOOL)flag;
@end

@implementation Brain

+ (void) initialize
{
    NSUserDefaults *defaults;
    NSMutableDictionary *fiendDefaults;
    
    if (self == [Brain class])
    {
        defaults = [NSUserDefaults standardUserDefaults];
        fiendDefaults = [NSMutableDictionary dictionary];

        // XXX: These two are not currently used.
        [fiendDefaults setObject:@"YES" forKey:DK_UseBoldDockFont];
        [fiendDefaults setObject:@"NO" forKey:DK_UseBigDockFont];

        // These are only used initially.
        [fiendDefaults setObject:@"0" forKey:DK_FiendTileScreen];
        [fiendDefaults setObject:@"-69" forKey:DK_FiendTileXOrigin];
        [fiendDefaults setObject:@"-69" forKey:DK_FiendTileYOrigin];

        [fiendDefaults setObject:@"NO" forKey:DK_DockLocked];
        [fiendDefaults setObject:@"NO" forKey:DK_DockSunken];

        // Thse are handled by the preferences panel
        [fiendDefaults setObject:NSStringFromFiendTileCorner (NorthEastCorner) forKey:DK_FiendTileCorner];
        [fiendDefaults setObject:@"YES" forKey:DK_PlaceTilesVertically];

        [fiendDefaults setObject:@"YES" forKey:DK_UseSounds];
        [fiendDefaults setObject:@"YES" forKey:DK_StrictCopy];
        [fiendDefaults setObject:@"NO" forKey:DK_HideTiles]; // Apps using the AnimatedAppTile framework don't show 'em by default.
        [fiendDefaults setObject:@"YES" forKey:DK_ShowLevelNumber];
        [fiendDefaults setObject:@"YES" forKey:DK_ClosePostLaunch];
        [fiendDefaults setObject:@"NO"  forKey:DK_EnableSingleClick];
        [fiendDefaults setObject:@"YES" forKey:DK_ShowFlagCues];

        [fiendDefaults setObject:@"/System/Library/CoreServices/Viewer.app" forKey:DK_WorkspacePath];

        [defaults registerDefaults:fiendDefaults];
    }
}

//----------------------------------------------------------------------

- init
{
    if ([super init] == nil)
        return nil;

    tileSettingsController = nil;

    dockWindow = nil;
    dockView = nil;

    saveTimer = nil;
    selectedTiles = [[NSMutableArray array] retain];

    //[self createDockWindow];

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    
    SNRelease (tileSettingsController);
    SNRelease (selectedTiles);

    // XXX: But only if saved?
    SNRelease (saveTimer);

    SNRelease (dockView);
    SNRelease (dockWindow);
    
    [super dealloc];
}

//----------------------------------------------------------------------

- (void) applicationDidFinishLaunching:(NSNotification *)aNotification
{
    NSWorkspace *sharedWorkspace;
    NSNotificationCenter *workspaceCenter;

    NSFileManager *fileManager;

    fileManager = [NSFileManager defaultManager];

    [fileManager createDirectoryAtPath:[@"~/Library" stringByStandardizingPath] attributes:nil];
    [fileManager createDirectoryAtPath:[@"~/Library/Fiend" stringByStandardizingPath] attributes:nil];

    dock = [FIDock instance];
    [[NSNotificationCenter defaultCenter] addObserver:self
                                          selector:@selector (dockLevelsDidChange:)
                                          name:FIDockLevelsDidChangeNotification
                                          object:dock];
    [dock setDelegate:self];

    [self createDockWindow];

    [self loadDock];
    [self showDock];
    [self rebuildLevelsMenu];

    sharedWorkspace = [NSWorkspace sharedWorkspace];
    workspaceCenter = [sharedWorkspace notificationCenter];

    [workspaceCenter addObserver:self
                     selector:@selector (workspaceNotification:)
                     name:nil
                     object:sharedWorkspace];

    //NSLog (@"active applications: %@", [FIAppInfo activeAppInfo]);

    // The autolaunching doesn't generate an applicationWillLaunch: notification.
    [dock performSelector:@selector (performAutoLaunch:) withObject:self afterDelay:0];

    [NSApp setServicesProvider:self];
}

//----------------------------------------------------------------------

- (void) applicationWillTerminate:(NSNotification *)aNotification
{
    [self saveDockIfNeeded];
    if ([[NSUserDefaults standardUserDefaults] boolForKey:DK_HideTiles] == YES)
    {
        // If we hid the tiles originally, we need to replace them when we quit.
        [dock unhideSourceTiles];
    }
}

//----------------------------------------------------------------------

- (void) workspaceNotification:(NSNotification *)aNotification
{
    //NSLog (@"notification: %@", aNotification);
    [[NSNotificationCenter fixedWorkspaceCenter] postNotification:aNotification];
}

//----------------------------------------------------------------------

- (void) showInfoPanel:sender
{
    [[InfoPanelController sharedInfoPanelController] showWindow:sender];
}

//----------------------------------------------------------------------

- (void) showPreferencePanel:sender
{
    [[PreferencePanelController sharedPreferencePanelController] showWindow:sender];
}

//----------------------------------------------------------------------

- (void) showActiveApplications:sender
{
    [[ActiveApplicationsController sharedActiveApplicationsController] showWindow:sender];
}

//----------------------------------------------------------------------

- (void) goToLevel:sender
{
    [dock goToLevelAtIndex:[sender tag]];
}

//----------------------------------------------------------------------
#if 0
- (void) findLevel:sender
{
}

//----------------------------------------------------------------------

- (void) findIcon:sender
{
}
#endif
//----------------------------------------------------------------------

- (void) addWorkspaceIcon:sender
{
    NSUserDefaults *defaults;

    defaults = [NSUserDefaults standardUserDefaults];
    [self addApplicationsToCurrentLevel:[NSArray arrayWithObject:[defaults objectForKey:DK_WorkspacePath]]];
}

//----------------------------------------------------------------------

- (void) placeLevelFirst:sender
{
    [dock placeCurrentLevelFirst];
}

//----------------------------------------------------------------------

- (void) placeLevelLast:sender
{
    [dock placeCurrentLevelLast];
}

//----------------------------------------------------------------------
#if 0
- (void) moveLevel:sender
{
}
#endif
//----------------------------------------------------------------------

- (void) rotateLevel:sender
{
    [dock rotateCurrentLevelAroundDockOrigin:[dockWindow frame].origin];
}

//----------------------------------------------------------------------

- (void) nameLevel:sender
{
    [[NameLevelController sharedNameLevelController] showWindow:sender];
}

//----------------------------------------------------------------------

- (void) addLevel:sender
{
    [dock addLevel];
    [self rebuildLevelsMenu];
}

//----------------------------------------------------------------------

- (void) clearLevel:sender
{
    [dock clearLevel];
}

//----------------------------------------------------------------------

- (void) deleteLevel:sender
{
    [dock removeLevel];
    [self rebuildLevelsMenu];
}

//----------------------------------------------------------------------
#if 0
- (void) moveIcons:sender
{
}
#endif
//----------------------------------------------------------------------

- (void) inspectSelectedIcon:sender
{
    if (tileSettingsController == nil)
    {
        tileSettingsController = [[TileSettingsController alloc] initWithDock:dock];
    }

    if ([selectedTiles count] > 0)
    {
        [tileSettingsController showWindowForTile:[selectedTiles objectAtIndex:0]];
    }
    else
    {
        NSBeep ();
    }
}

//----------------------------------------------------------------------

- (void) showReleaseNotes:sender
{
    NSString *path;

    path = [[NSBundle mainBundle] pathForResource:@"ac_ReleaseNotes" ofType:@"rtf"];
    [[NSWorkspace sharedWorkspace] openFile:path];
}
    
//----------------------------------------------------------------------

- (void) showInstructions:sender
{
    NSString *path;

    path = [[NSBundle mainBundle] pathForResource:@"ad_Instructions" ofType:@"rtfd"];
    [[NSWorkspace sharedWorkspace] openFile:path];
}

//----------------------------------------------------------------------

- (void) go:sender
{
    NSLog (@"[%@]levelsMenuItem: %@", NSStringFromClass ([levelsMenuItem class]), levelsMenuItem);
    //NSLog (@"active applications: %@", [FIAppInfo activeAppInfo]);
    //[dock saveDock];
}

//----------------------------------------------------------------------
// NSApplication delegate
//----------------------------------------------------------------------

- (void) applicationDidBecomeActive:(NSNotification *)aNotification
{
}

//----------------------------------------------------------------------

- (void) applicationWillBecomeActive:(NSNotification *)aNotification
{
    // This seems to fix the problem of starting Fiend, double-clicking Stuart, and then
    // trying to activate Fiend (The menu didn't show up, and ...DidBecomeActive was never called).
    [NSApp unhide:self];
}

//----------------------------------------------------------------------
// Document handling
//----------------------------------------------------------------------

- (NSString *) screenSizeString
{
    NSScreen *screen;
    NSSize screenSize;

    screen = [[NSScreen screens] objectAtIndex:0];
    screenSize = [screen frame].size;

    return [NSString stringWithFormat:@"%.0f.%.0f", screenSize.width, screenSize.height];
}

//----------------------------------------------------------------------

- (NSString *) dockFilename
{
    return [NSString stringWithFormat:@"FiendDock.%@", [self screenSizeString]];
}

//----------------------------------------------------------------------

- (NSString *) dockPath
{
    return [NSString stringWithFormat:@"%@/Library/Fiend/%@", NSHomeDirectory (), [self dockFilename]];
}

//----------------------------------------------------------------------

- (void) loadDock
{
    NSString *dockPath;
    NSDictionary *savedDock;

    dockPath = [self dockPath];
    //NSLog (@"dockPath: %@", dockPath);
    savedDock = [NSDictionary dictionaryWithContentsOfFile:dockPath];
    if (savedDock == nil)
    {
        NSLog (@"Could not load dock from: %@", dockPath);
        return;
    }
    //NSLog (@"savedDock: %@", savedDock);

    [dock loadFromPropertyList:savedDock];
    [dock repositionTilesRelativeToPoint:[dockWindow frame].origin];


    // Check for already active applications
    [dock checkForActiveApplications:[FIAppInfo activeAppInfo]];

    // Perform autolaunching of applications.
    //[self performAutoLaunch:self];
}

//----------------------------------------------------------------------

- (void) saveDockIfNeeded
{
    if (saveTimer != nil)
    {
        [self stopSaveTimer];
        [self saveDock];
    }
}

//----------------------------------------------------------------------

- (void) saveDock
{
    NSString *dockPath;
    NSDictionary *savedDock;
    BOOL savedOkay;
    
    dockPath = [self dockPath];

    savedDock = [dock propertyListForDock];

    [[NSFileManager defaultManager] copyPath:dockPath toPath:[NSString stringWithFormat:@"%@~", dockPath] handler:nil];
    savedOkay = [savedDock writeToFile:dockPath atomically:YES];
    if (savedOkay == NO)
    {
        NSLog (@"Could not save dock to: %@", dockPath);
        NSBeep ();
    }
}

//----------------------------------------------------------------------

- (void) createDockWindow
{
    NSScreen *mainScreen;
    NSRect frame;
    NSRect mainScreenFrame;

    mainScreen = [NSScreen mainScreen];
    
    mainScreenFrame = [mainScreen frame];

    frame.origin = [self initialFiendTilePoint];
    frame.size = NSMakeSize (STANDARD_TILE_WIDTH, STANDARD_TILE_HEIGHT);

    // Note: It appears that the first responder border is only shown around, for example, buttons, when the window
    //       has a title bar...
    dockWindow = [[FIWindow alloc] initWithContentRect:frame
                                   styleMask:NSBorderlessWindowMask
                                   backing:NSBackingStoreBuffered
                                   defer:NO];

    [dockWindow setLevel:[dock tileWindowLevel]];
    [dockWindow _setAvoidsActivation:YES];
    [dockWindow setFrameAutosaveName:[NSString stringWithFormat:@"DockWindow.%@", [self screenSizeString]]];

    dockView = [[FIDockView alloc] initWithFrame:NSMakeRect (0, 0, STANDARD_TILE_WIDTH, STANDARD_TILE_HEIGHT)];
    [dockWindow setContentView:dockView];
    [dockView setDatasource:dock];
    [dockView setDelegate:self];
    [dockView release];

    [dockWindow setDelegate:self];
    //[dockWindow orderFront:self];
    //[dockWindow release];
}

//----------------------------------------------------------------------

- (void) showDock
{
    [dockWindow orderFront:self];
    [dock unhide];
}

//----------------------------------------------------------------------
// Save timer
//----------------------------------------------------------------------

- (void) startSaveTimer
{
    [self stopSaveTimer];

    saveTimer = [[NSTimer scheduledTimerWithTimeInterval:30
                          target:self
                          selector:@selector (fireSaveTimer:)
                          userInfo:nil
                          repeats:NO] retain];
}

//----------------------------------------------------------------------

- (void) stopSaveTimer
{
    if (saveTimer != nil)
    {
        [saveTimer invalidate];
        SNRelease (saveTimer);
    }
}

//----------------------------------------------------------------------

- (void) fireSaveTimer:(NSTimer *)timer
{
    [self stopSaveTimer]; // Just in case.
    [self saveDock];
}

//----------------------------------------------------------------------
// FIDock delegate
//----------------------------------------------------------------------

- (void) updateChangeCount
{
    [self startSaveTimer];
    [dockView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (void) updateDockView
{
    [dockView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------
// FIDockView Delegate Protocol
//----------------------------------------------------------------------

- (void) nextLevel:sender
{
    [dock nextLevel];
    // Otherwise, it doesn't get updated. XXX: double check
    [dockView display];
}

//----------------------------------------------------------------------

- (void) previousLevel:sender
{
    [dock previousLevel];
    // Otherwise, it doesn't get updated. XXX: double check
    [dockView display];
}

//----------------------------------------------------------------------

- (void) activateApplication:sender
{
    // We don't actually have to do anything - just clicking the button
    // will activate the application.  Of course, clicking the next/previous
    // level buttons also does.
#if 0
    DSTART;
    NSLog (@"isActive: %@", ([NSApp isActive] == YES) ? @"Yes" : @"No");
    //[NSApp deactivate];
    [NSApp unhide:self];
    [NSApp activateIgnoringOtherApps:YES];
    //[NSApp hide:self];
    //[NSApp unhide:self];
    DEND;
#endif
}

//----------------------------------------------------------------------

- (void) addApplicationsToCurrentLevel:(NSArray *)applications
{
    NSEnumerator *filenameEnumerator;
    NSString *filename;
    FITile *tile;
    DockLocation dockLocation;
    NSDictionary *activeApplications;
    FILevel *currentLevel;
    NSPoint dockTileOrigin;

    activeApplications = [FIAppInfo activeAppInfo];
    currentLevel = [dock currentLevel];
    dockTileOrigin = [dockWindow frame].origin;

    filenameEnumerator = [applications objectEnumerator];
    while (filename = [filenameEnumerator nextObject])
    {
        if ([filename isApplicationPath] == YES)
        {
            if ([dock containsApplication:filename] == NO)
            {
                // Get next available icon position (based on current & persistent dockLevels)
                tile = [[[FITile alloc] initWithRepresentedPath:filename] autorelease];
                dockLocation = [dock nextAvailableDockLocation:dockTileOrigin];
                [tile setLocation:dockLocation];
                [tile createTileWindow];
                [tile repositionRelativeToPoint:[[dockView window] frame].origin];
                [tile setTileWindowLevel:[dock tileWindowLevel]];
                [tile unhide];
                [currentLevel addTile:tile];
                [tile checkForActiveApplication:activeApplications];
                [self updateChangeCount];
            }
            else
            {
                NSLog (@"Application is already on the dock: %@", filename);
            }
        }
        else
        {
            NSLog (@"File is not an application (.app, .debug, or .profile): %@", filename);
        }
    }

    // Reposition tiles
    // unhide them
}

//----------------------------------------------------------------------

// XXX: Should we have the menu item show "Show Dock"/"Hide Dock"?

- (void) toggleDockVisible:sender
{
    [dock setIsVisible:![dock isVisible]];
    if ([dock isVisible] == YES)
        [dock unhide];
    else
        [dock hide];
}

//----------------------------------------------------------------------

- (void) toggleDockSunken:sender
{
    [dock setIsSunken:![dock isSunken]];
    [dockWindow setLevel:[dock tileWindowLevel]];
    [dock resetWindowLevels];
}

//----------------------------------------------------------------------

// XXX: Should we have the menu item show "Lock Dock"/"Unlock Dock"?
- (void) toggleDockLocked:sender
{
    [dock setIsLocked:![dock isLocked]];
}

//----------------------------------------------------------------------

- (void) dockWindowDidDrag:(NSPoint)newWindowOrigin
{
    [dock repositionVisibleTilesRelativeToPoint:newWindowOrigin];
}

//----------------------------------------------------------------------
// NSWindow delegate
//----------------------------------------------------------------------

- (void) windowDidMove:(NSNotification *)aNotification
{
    NSPoint newOrigin;

    newOrigin = [[dockView window] frame].origin;
    [dock repositionTilesRelativeToPoint:newOrigin];
}

//----------------------------------------------------------------------
// Selection
//----------------------------------------------------------------------

- (void) selectTile:(FITile *)tile
{
    [tile setIsSelected:YES];
    [selectedTiles addObject:tile];
}

//----------------------------------------------------------------------

// Select all tiles on current and persistent levels.

- (void) selectAll:sender
{
    NSEnumerator *tileEnumerator;
    FITile *tile;

    tileEnumerator = [[[dock persistentLevel] tiles] objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        [self selectTile:tile];
    }

    tileEnumerator = [[[dock currentLevel] tiles] objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        [self selectTile:tile];
    }
}

//----------------------------------------------------------------------

- (void) deselectTile:(FITile *)tile
{
    [tile setIsSelected:NO];
    [selectedTiles removeObject:tile];
}

//----------------------------------------------------------------------

- (void) toggleSelection:sender
{
    FITile *tile;

    if ([sender isKindOfClass:[FITile class]] == YES)
    {
        tile = sender;

        if ([tile isSelected] == YES)
            [self deselectTile:tile];
        else
            [self selectTile:tile];
    }
}

//----------------------------------------------------------------------

- (void) deleteSelectedTiles
{
    [dock deleteTiles:selectedTiles];
    [selectedTiles removeAllObjects];
}

//----------------------------------------------------------------------

- (void) deselectAllTiles:sender
{
    NSEnumerator *tileEnumerator;
    FITile *tile;

    tileEnumerator = [selectedTiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        [tile setIsSelected:NO];
    }

    [selectedTiles removeAllObjects];
}

//----------------------------------------------------------------------
// Cut & Paste
//----------------------------------------------------------------------

- (void) cut:sender
{
    [self copy:sender];
    [self deleteSelectedTiles];
    [[FISoundManager sharedSoundManager] didRemoveTiles];
}

//----------------------------------------------------------------------

- (void) copy:sender
{
    NSPasteboard *generalPasteboard;
    NSMutableArray *applicationPaths;
    NSEnumerator *tileEnumerator;
    FITile *tile;
    int changeCount;

    generalPasteboard = [NSPasteboard generalPasteboard];
    applicationPaths = [NSMutableArray array];
    tileEnumerator = [selectedTiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        [applicationPaths addObject:[tile path]];
    }
    changeCount = [generalPasteboard declareTypes:[NSArray arrayWithObject:NSFilenamesPboardType] owner:nil];
    //NSLog (@"applicationPaths: %@", applicationPaths);
    if ([generalPasteboard setPropertyList:applicationPaths forType:NSFilenamesPboardType] == NO)
    {
        NSLog (@"setPropertyList:forType: returned NO.");
    }
}

//----------------------------------------------------------------------

- (void) paste:sender
{
    NSPasteboard *generalPasteboard;
    NSString *availableType;
    NSArray *filenames;

    generalPasteboard = [NSPasteboard generalPasteboard];
    availableType = [generalPasteboard availableTypeFromArray:[NSArray arrayWithObject:NSFilenamesPboardType]];

    if (availableType == NSFilenamesPboardType)
    {
        filenames = [generalPasteboard propertyListForType:NSFilenamesPboardType];

        [self addApplicationsToCurrentLevel:filenames];
        [[FISoundManager sharedSoundManager] didAddTiles];
    }
    else
    {
        NSBeep ();
    }
}

//----------------------------------------------------------------------
// Other
//----------------------------------------------------------------------

// It would be nice if we could used X-style negative values to indicate,
// for example, the right edge of the tile being -N pixels from the right
// edge.  The problem with this is distinguishing -0 from 0.

- (NSPoint) initialFiendTilePoint
{
    NSUserDefaults *defaults;
    int screenNumber, screenCount;
    NSSize screenSize;
    NSPoint fiendTileOrigin;
    NSArray *screens;

    defaults = [NSUserDefaults standardUserDefaults];
    screens = [NSScreen screens];
    screenCount = [screens count];

    screenNumber = [defaults integerForKey:DK_FiendTileScreen];
    if (screenNumber >= screenCount)
    {
        NSLog (@"Default screen number (%d) beyond total number of screens (%d).", screenNumber, screenCount);
        screenNumber = 0;
    }

    fiendTileOrigin.x = [defaults floatForKey:DK_FiendTileXOrigin];
    fiendTileOrigin.y = [defaults floatForKey:DK_FiendTileYOrigin];

    screenSize = [[screens objectAtIndex:screenNumber] visibleFrame].size;

    if (fiendTileOrigin.x < 0)
        fiendTileOrigin.x += screenSize.width /*- STANDARD_TILE_WIDTH*/;
    if (fiendTileOrigin.y < 0)
        fiendTileOrigin.y += screenSize.height /*- STANDARD_TILE_HEIGHT*/;

    return fiendTileOrigin;
}

//----------------------------------------------------------------------

- (BOOL) validateMenuItem:(NSMenuItem *)menuCell
{
    SEL action = [menuCell action];
    BOOL valid = YES;

    if (action == @selector (cut:)
        || action == @selector (copy:))
    {
        valid = [selectedTiles count] > 0;
    }

    return valid;
}

//----------------------------------------------------------------------

- (void) rebuildLevelsMenu
{
    NSArray *levelIdentifiers;
    NSEnumerator *levelNameEnumerator;
    NSString *levelName;
    int index;
    NSMenuItem *menuItem;
    id oldTarget;

    oldTarget = [levelsMenuItem target];

    while (menuItem = [[oldTarget itemArray] lastObject])
    {
        [oldTarget removeItem:menuItem];
    }

    index = 0;
    levelIdentifiers = [dock levelIdentifiers];
    levelNameEnumerator = [levelIdentifiers objectEnumerator];
    while (levelName = [levelNameEnumerator nextObject])
    {
        menuItem = [oldTarget insertItemWithTitle:levelName
                                  action:@selector (goToLevel:)
                                  keyEquivalent:@""
                                  atIndex:index];
        [menuItem setTag:index];
        index++;
    }
}

//----------------------------------------------------------------------

- (void) dockLevelsDidChange:(NSNotification *)aNotification
{
    [self rebuildLevelsMenu];
}

//----------------------------------------------------------------------
// Services
//----------------------------------------------------------------------

- (void) submitToFiendDock:(NSPasteboard *)pasteboard userData:(NSData *)data error:(NSString **)error
{
    NSArray *filenames;

    if ([pasteboard availableTypeFromArray:[NSArray arrayWithObject:NSFilenamesPboardType]] != nil)
    {
        filenames = [pasteboard propertyListForType:NSFilenamesPboardType];
        [self addApplicationsToCurrentLevel:filenames];
    }
}

@end
