#import "Fiend.h"

RCSID ("$Id: FIAppInfo.m,v 1.1.1.1 1998/09/03 01:14:53 nygard Exp $");

#import "FIAppInfo.h"

#import <AppKit/psopsNeXT.h>
#import "FiendWraps.h"
#import "table.h"
#import "Fiendish.h"

#include <sys/types.h>
#include <pwd.h>

#define WMAX (1000)

@implementation FIAppInfo

+ (NSDictionary *) activeAppInfo
{
    NSMutableSet *set;
    NSMutableDictionary *dictionary;
    int l, count;
    int windowNumbers[WMAX];
    int anOwnerContext;
    FIAppInfo *info;
    NSEnumerator *numberEnumerator;
    NSNumber *number;

    set = [NSMutableSet set];

    PScountwindowlist (0, &count);
    if (count > WMAX)
        count = WMAX;

    PSwindowlist (0, count, windowNumbers);
    for (l = 0; l < count; l++)
    {
        if (safeGetCurrentOwner (windowNumbers[l], &anOwnerContext) == 0)
        {
            [set addObject:[NSNumber numberWithInt:anOwnerContext]];
        }
    }

    //NSLog (@"anOwnerContexts: %@", set);

    dictionary = [NSMutableDictionary dictionary];
    numberEnumerator = [set objectEnumerator];
    while (number = [numberEnumerator nextObject])
    {
        info = [FIAppInfo appInfoWithOwnerContext:[number intValue]];
        [dictionary setObject:info forKey:[info shortName]];
    }

    //NSLog (@"active applications: %@", dictionary);

    return dictionary;
}

//----------------------------------------------------------------------

+ appInfoWithOwnerContext:(int)anOwnerContext
{
    FIAppInfo *info;
    int aPid, aTokenWindowNumber;
    struct tbl_procinfo process_info;

    info = nil;
    if (safeGetTokenWindowAndPid (anOwnerContext, &aTokenWindowNumber, &aPid) != -1)
    {
        if (table (TBL_PROCINFO, aPid, &process_info, 1, sizeof (process_info)) == 1
            && process_info.pi_status == PI_ACTIVE)
        {
            info = [[[FIAppInfo alloc] initWithOwnerContext:anOwnerContext
                                       tokenWindowNumber:aTokenWindowNumber
                                       pid:aPid
                                       shortName:[NSString stringWithCString:process_info.pi_comm]
                                       uid:process_info.pi_uid] autorelease];
        }
    }
    else
    {
        NSLog (@"safeGetTokenWindowAndPid() failed.");
    }

    return info;
}

//----------------------------------------------------------------------

- initWithOwnerContext:(int)anOwnerContext
     tokenWindowNumber:(int)aWindowNumber
                   pid:(int)aPid
             shortName:(NSString *)aShortName
                   uid:(int)aUid
{
    if ([super init] == nil)
        return nil;

    ownerContext = anOwnerContext;
    tokenWindowNumber = aWindowNumber;
    pid = aPid;
    shortName = [aShortName copy];

    uid = aUid;
    userName = nil; // Create lazily.

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    SNRelease (shortName);

    [super dealloc];
}

//----------------------------------------------------------------------

- (int) ownerContext
{
    return ownerContext;
}

//----------------------------------------------------------------------

- (int) pid
{
    return pid;
}

//----------------------------------------------------------------------

- (int) tokenWindowNumber
{
    return tokenWindowNumber;
}

//----------------------------------------------------------------------

- (NSString *) shortName
{
    return shortName;
}

//----------------------------------------------------------------------

- (int) uid
{
    return uid;
}

//----------------------------------------------------------------------

- (NSString *) userName
{
    if (userName == nil)
    {
        struct passwd *pw;

        pw = getpwuid (uid);
        if (pw != NULL)
        {
            userName = [[NSString stringWithCString:pw->pw_name] retain];
        }
    }

    return userName;
}

//----------------------------------------------------------------------

- (NSString *) description
{
    return [NSString stringWithFormat:@"<FIAppInfo: %d, %d, %d, '%@'>",
                     ownerContext, tokenWindowNumber, pid, shortName];
}

//----------------------------------------------------------------------

- (void) activateApplicationAndHideOthers:(BOOL)hideOthers
{
    int status;
    
    status = safeUnhideContext (ownerContext, YES, YES, hideOthers);
    if (status != 0)
        NSLog (@"safeUnhideContext() failed.");
}

@end
