//
// $Id: FIDock.h,v 1.1.1.1 1998/09/03 01:14:52 nygard Exp $
//

#import <Foundation/Foundation.h>

@class FILevel;
@class FIDockView;
@class FITile;

extern NSString *FIDockLevelsDidChangeNotification;

@interface FIDock : NSObject
{
    FILevel *persistentLevel;
    NSMutableArray *levels;

    FILevel *currentLevel;

    BOOL isVisible;

    id nonretainedDelegate;
}

+ instance;

+ (void) initialize;

- init;
- (void) dealloc;

- (void) loadFromPropertyList:(NSDictionary *)savedDockPropertyList;
- (NSDictionary *) propertyListForDock;

- delegate;
- (void) setDelegate:newDelegate;

- (FILevel *) persistentLevel;

- (FILevel *) currentLevel;
- (void) setCurrentLevel:(FILevel *)newLevel;

- (BOOL) isVisible;
- (void) setIsVisible:(BOOL)newFlag;

- (BOOL) isLocked;
- (void) setIsLocked:(BOOL)newFlag;

- (BOOL) isSunken;
- (void) setIsSunken:(BOOL)newFlag;

- (int) tileWindowLevel;
- (void) resetWindowLevels;

- (void) addLevel;
- (void) removeLevel;

- (void) nextLevel;
- (void) previousLevel;

- (void) placeCurrentLevelFirst;
- (void) placeCurrentLevelLast;

- (void) clearLevel;
- (void) rotateCurrentLevelAroundDockOrigin:(NSPoint)origin;
- (void) deleteTiles:(NSArray *)tiles;

//----------------------------------------------------------------------
// FIDockView datasource
//----------------------------------------------------------------------

- (NSString *) currentLevelNumber;
- (NSString *) currentLevelName;
//- (BOOL) isVisible;
//- (BOOL) isLocked;

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------

- (DockLocation) nextAvailableDockLocation:(NSPoint)dockTileOrigin;
- (DockLocation) nextAvailableDockLocation:(NSPoint)dockTileOrigin orLocation:(DockLocation)aLocation;
- (DockLocation) nextAvailableDockLocation2:(NSPoint)dockTileOrigin orLocation:(DockLocation)aLocation;
- (BOOL) isLocationAvailable:(DockLocation)aLocation;

- (BOOL) containsApplication:(NSString *)applicationPath;

//----------------------------------------------------------------------
// For the GoToLevel panel
//----------------------------------------------------------------------

- (NSArray *) levelIdentifiers;
- (int) indexOfCurrentLevel;
- (void) goToLevelAtIndex:(int)index;

//----------------------------------------------------------------------
// NSWorkspace notifications
//----------------------------------------------------------------------

- (void) applicationWillLaunch:(NSNotification *)aNotification;
- (void) applicationDidLaunch:(NSNotification *)aNotification;
- (void) applicationDidTerminate:(NSNotification *)aNotification;

//----------------------------------------------------------------------
// Fiend Launcher notifications
//----------------------------------------------------------------------

- (void) applicationLaunchFailed:(NSNotification *)aNotification;

//----------------------------------------------------------------------
// Preference notifications
//----------------------------------------------------------------------

- (void) hideTilesDidChange:(NSNotification *)aNotification;
- (void) showLevelNumberDidChange:(NSNotification *)aNotification;
- (void) showFlagCuesDidChange:(NSNotification *)aNotification;

//----------------------------------------------------------------------
// Other
//----------------------------------------------------------------------

- (void) unhide;
- (void) hide;
- (void) checkForActiveApplications:(NSDictionary *)activeApplications;
- (void) performAutoLaunch:sender;
- (void) hideSourceTiles;
- (void) unhideSourceTiles;
- (void) updateTileViews;

- (void) updateChangeCount;
- (void) updateDockView;
- (void) levelsDidChange;

- (void) repositionTilesRelativeToPoint:(NSPoint)newOrigin;
- (void) repositionVisibleTilesRelativeToPoint:(NSPoint)newOrigin;

- (void) togglePersistenceOfTile:(FITile *)aTile;

@end
