#import "Fiend.h"

RCSID ("$Id: FIDock.m,v 1.1.1.1 1998/09/03 01:14:51 nygard Exp $");

#import "FIDock.h"

#import "FILevel.h"
#import "FITile.h"
#import "FIDockView.h"

#import "FIWindow.h"
#import "NSNotificationCenter+FIExtensions.h"
#import "FIAppInfo.h"
#import "FiendDefaults.h"
#import "FILauncher.h"

#import "FiendWraps.h"

#define TILE_WIDTH (64)
#define TILE_HEIGHT (64)
#define DOCK_RIGHT_MARGIN (3)
#define DOCK_TOP_MARGIN (0)

@interface NSApplication (Private)
- _appIcon;
@end

//======================================================================

DEFINE_NSSTRING (FIDockLevelsDidChangeNotification);

@implementation FIDock

+ instance
{
    static FIDock *_instance = nil;

    if (_instance == nil)
    {
        _instance = [[FIDock alloc] init];
    }

    return _instance;
}

//----------------------------------------------------------------------

+ (void) initialize
{
}

//----------------------------------------------------------------------

- init
{
    NSWorkspace *sharedWorkspace;
    NSNotificationCenter *workspaceCenter;
    NSNotificationCenter *defaultCenter;

    if ([super init] == nil)
        return nil;

    persistentLevel = [[FILevel alloc] initWithDock:self];
    [persistentLevel setIsPersistent:YES];
    
    levels = [[NSMutableArray array] retain];

    currentLevel = [[FILevel alloc] initWithDock:self];
    [levels addObject:currentLevel];

    isVisible = YES; // Put before creating dock view.

    sharedWorkspace = [NSWorkspace sharedWorkspace];
    // Adding this as an observer appears to break the observation of the brain...
    //workspaceCenter = [sharedWorkspace notificationCenter];
    workspaceCenter = [NSNotificationCenter fixedWorkspaceCenter];

    [workspaceCenter addObserver:self
                     selector:@selector (applicationWillLaunch:)
                     name:NSWorkspaceWillLaunchApplicationNotification
                     object:sharedWorkspace];

    [workspaceCenter addObserver:self
                     selector:@selector (applicationDidLaunch:)
                     name:NSWorkspaceDidLaunchApplicationNotification
                     object:sharedWorkspace];

    [workspaceCenter addObserver:self
                     selector:@selector (applicationDidTerminate:)
                     name:NSWorkspaceDidTerminateApplicationNotification
                     object:sharedWorkspace];

    defaultCenter = [NSNotificationCenter defaultCenter];
    [defaultCenter addObserver:self
                   selector:@selector (hideTilesDidChange:)
                   name:FIPreferencesHideTilesDidChangeNotification
                   object:nil];

    [defaultCenter addObserver:self
                   selector:@selector (showLevelNumberDidChange:)
                   name:FIPreferencesShowLevelNumberDidChangeNotification
                   object:nil];

    [defaultCenter addObserver:self
                   selector:@selector (showFlagCuesDidChange:)
                   name:FIPreferencesShowFlagCuesDidChangeNotification
                   object:nil];

    [defaultCenter addObserver:self
                   selector:@selector (applicationLaunchFailed:)
                   name:FIFailedToLaunchApplicationNotificaiton
                   object:nil];

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    //[[[NSWorkspace sharedWorkspace] notificationCenter] removeObserver:self];
    [[NSNotificationCenter fixedWorkspaceCenter] removeObserver:self];
    SNRelease (persistentLevel);
    SNRelease (currentLevel);
    SNRelease (levels);

    [super dealloc];
}

//----------------------------------------------------------------------

// XXX: Change this into an init method.
- (void) loadFromPropertyList:(NSDictionary *)savedDockPropertyList
{
    NSDictionary *savedLevel;
    NSArray *loadedLevels;
    NSEnumerator *levelEnumerator;

    // NOTE: This should only be run on newly created docks.  It should
    //       probably be changed into an -init method.

    savedLevel = [savedDockPropertyList objectForKey:@"PersistentDockLevel"];
    if (savedLevel != nil)
    {
        [persistentLevel loadApplications:[savedLevel objectForKey:@"DockedApplications"]];
    }

    loadedLevels = [savedDockPropertyList objectForKey:@"DockLevels"];
    if (loadedLevels != nil)
    {
        BOOL isFirstLevel;
        FILevel *newLevel;

        isFirstLevel = YES;
        levelEnumerator = [loadedLevels objectEnumerator];
        while (savedLevel = [levelEnumerator nextObject])
        {
            if (isFirstLevel == YES)
            {
                newLevel = currentLevel;
            }
            else
            {
                newLevel = [[[FILevel alloc] initWithDock:self] autorelease];
                [levels addObject:newLevel];
            }
            
            [newLevel loadApplications:[savedLevel objectForKey:@"DockedApplications"]];
            [newLevel setLevelName:[savedLevel objectForKey:@"DockLevelName"]];
            isFirstLevel = NO;
        }
    }
}

//----------------------------------------------------------------------

- (NSDictionary *) propertyListForDock
{
    NSMutableDictionary *savedDock;
    NSMutableArray *savedLevels;
    NSEnumerator *levelEnumerator;
    FILevel *level;

    savedDock = [NSMutableDictionary dictionary];
    [savedDock setObject:[persistentLevel levelPropertyList] forKey:@"PersistentDockLevel"];

    savedLevels = [NSMutableArray array];
    levelEnumerator = [levels objectEnumerator];
    while (level = [levelEnumerator nextObject])
    {
        [savedLevels addObject:[level levelPropertyList]];
    }

    [savedDock setObject:savedLevels forKey:@"DockLevels"];

    return savedDock;
}

//----------------------------------------------------------------------

- delegate
{
    return nonretainedDelegate;
}

//----------------------------------------------------------------------

- (void) setDelegate:newDelegate
{
    if (newDelegate == nonretainedDelegate)
        return;

    nonretainedDelegate = newDelegate;
}

//----------------------------------------------------------------------

- (FILevel *) persistentLevel
{
    return persistentLevel;
}

//----------------------------------------------------------------------

- (FILevel *) currentLevel
{
    return currentLevel;
}

//----------------------------------------------------------------------

- (void) setCurrentLevel:(FILevel *)newLevel
{
    if (newLevel == currentLevel)
        return;

    NSAssert ([levels containsObject:newLevel] == YES, @"Expected current level to be in array of levels.");

    [currentLevel hideLevel];

    SNRelease (currentLevel);
    currentLevel = [newLevel retain];

    [self updateDockView];
    // But it (displaying) causes deleting a non-last level to "flicker" the level name.

    [persistentLevel setTileWindowLevel:[self tileWindowLevel]];
    [currentLevel setTileWindowLevel:[self tileWindowLevel]];
    isVisible = YES;
    [self unhide];
}

//----------------------------------------------------------------------

- (BOOL) isVisible
{
    return isVisible;
}

//----------------------------------------------------------------------

- (void) setIsVisible:(BOOL)newFlag
{
    if (newFlag == isVisible)
        return;

    isVisible = newFlag;

    [self updateDockView];
}

//----------------------------------------------------------------------

- (BOOL) isLocked
{
    return [[NSUserDefaults standardUserDefaults] boolForKey:DK_DockLocked];
}

//----------------------------------------------------------------------

- (void) setIsLocked:(BOOL)newFlag
{
    [[NSUserDefaults standardUserDefaults] setBool:newFlag forKey:DK_DockLocked];
    [self updateDockView];
}

//----------------------------------------------------------------------

- (BOOL) isSunken
{
    return [[NSUserDefaults standardUserDefaults] boolForKey:DK_DockSunken];
}

//----------------------------------------------------------------------

- (void) setIsSunken:(BOOL)newFlag
{
    NSUserDefaults *defaults;
    BOOL isSunken;

    defaults = [NSUserDefaults standardUserDefaults];
    isSunken = [defaults boolForKey:DK_DockSunken];
    if (newFlag == isSunken)
        return;

    [defaults setBool:newFlag forKey:DK_DockSunken];
}

//----------------------------------------------------------------------

- (int) tileWindowLevel
{
    return ([self isSunken] == YES) ? (NSNormalWindowLevel - 1) : NSDockWindowLevel;
}

//----------------------------------------------------------------------

- (void) resetWindowLevels
{
    [persistentLevel setTileWindowLevel:[self tileWindowLevel]];
    [currentLevel setTileWindowLevel:[self tileWindowLevel]];
}

//----------------------------------------------------------------------

- (void) addLevel
{
    FILevel *newLevel;

    newLevel = [[[FILevel alloc] initWithDock:self] autorelease];
    [levels addObject:newLevel];
    [self goToLevelAtIndex:[levels count] - 1];
    [self updateChangeCount];
}

//----------------------------------------------------------------------

- (void) removeLevel
{
    int currentLevelIndex, levelCount;
    FILevel *targetLevel;
    int tileCount;
    int returnValue;
    
    targetLevel = nil;
    levelCount = [levels count];
    if (levelCount > 1)
    {
        tileCount = [[currentLevel tiles] count];
        if (tileCount > 0)
        {
            returnValue = NSRunAlertPanel (@"Delete Level",
                                           @"You are about to remove a dock level with %d non-persistent application%@.",
                                           @"Cancel", @"Remove", nil, tileCount, tileCount == 1 ? @"" : @"s");
            if (returnValue != NSAlertAlternateReturn)
                return;
        }

        targetLevel = currentLevel;
        currentLevelIndex = [levels indexOfObject:currentLevel];
                
        if (currentLevelIndex + 1 < levelCount)
        {
            [self goToLevelAtIndex:currentLevelIndex + 1];
        }
        else
        {
            [self goToLevelAtIndex:currentLevelIndex - 1];
        }

        [levels removeObject:targetLevel];

        [self updateChangeCount];
    }
    else
    {
        NSBeep ();
    }
}

//----------------------------------------------------------------------

- (void) nextLevel
{
    int count;
    int current;

    count = [levels count];
    if (count < 2)
        return;

    current = [levels indexOfObject:currentLevel];
    current = (current + 1) % count;
    [self goToLevelAtIndex:current];
}

//----------------------------------------------------------------------

- (void) previousLevel
{
    int count;
    int current;

    count = [levels count];
    if (count < 2)
        return;

    current = [levels indexOfObject:currentLevel];
    if (current == 0)
        current = count - 1;
    else
        current--;

    [self goToLevelAtIndex:current];
}

//----------------------------------------------------------------------

- (void) placeCurrentLevelFirst
{
    [levels removeObject:currentLevel];
    [levels insertObject:currentLevel atIndex:0];
    [self updateChangeCount];
    [self levelsDidChange];
}

//----------------------------------------------------------------------

- (void) placeCurrentLevelLast
{
    [levels removeObject:currentLevel];
    [levels addObject:currentLevel];
    [self updateChangeCount];
    [self levelsDidChange];
}

//----------------------------------------------------------------------

- (void) clearLevel
{
    int tileCount;
    int returnValue;

    tileCount = [[currentLevel tiles] count];
    if (tileCount > 0)
    {
        returnValue = NSRunAlertPanel (@"Clear Level",
                                       @"You are about to remove %d non-persistent application%@ from the current dock level.",
                                       @"Cancel", @"Remove", nil, tileCount, tileCount == 1 ? @"" : @"s");
        if (returnValue == NSAlertAlternateReturn)
        {
            [currentLevel clearLevel];
            [self updateChangeCount];
        }
    }
}

//----------------------------------------------------------------------

- (void) rotateCurrentLevelAroundDockOrigin:(NSPoint)origin
{
    [currentLevel rotateTilesAroundPoint:origin];
    [self updateChangeCount];
}

//----------------------------------------------------------------------

- (void) deleteTiles:(NSArray *)tiles
{
    [persistentLevel deleteTiles:tiles];
    [levels makeObjectsPerformSelector:@selector (deleteTiles:) withObject:tiles];
    [self updateChangeCount]; // XXX: But only if some were removed? ([tiles count] > 0)
}

//----------------------------------------------------------------------
// FIDockView datasource
//----------------------------------------------------------------------

- (NSString *) currentLevelNumber
{
    return [NSString stringWithFormat:@"%d", [levels indexOfObject:currentLevel] + 1];
}

//----------------------------------------------------------------------

- (NSString *) currentLevelName
{
    NSString *levelName;

    levelName = [currentLevel levelName];
    if (levelName == nil)
    {
        levelName = [NSString stringWithFormat:@"%d-%d",
                              [levels indexOfObject:currentLevel] + 1,
                              [levels count]];
    }

    return levelName;
}

//----------------------------------------------------------------------
//
//----------------------------------------------------------------------

// (0,0) is reserved for the FIDockView, and will be used to
// represent "no available location"

// This has to search both the current dock and the persistent dock.
// This searching seems to be fast enough, especially for the normal
// numbers and frequencies...

- (DockLocation) nextAvailableDockLocation:(NSPoint)dockTileOrigin
{
    DockLocation location;

    location.x = -1000;
    location.y = -1000;
    return [self nextAvailableDockLocation:dockTileOrigin orLocation:location];
}

//----------------------------------------------------------------------

//----------------------------------------------------------------------

- (DockLocation) nextAvailableDockLocation:(NSPoint)dockTileOrigin orLocation:(DockLocation)aLocation
{
    int minx, miny, maxx, maxy;
    DockLocation current;
    NSRect mainScreenFrame;
    NSUserDefaults *defaults;
    NSString *fiendTileCorner;
    BOOL placeTilesVertically;

    int endX, endY, deltaX, deltaY;

    // Calculate number of tiles the will fit below, left, above, and right of the main tile.
    // (Calculate total number of tiles that will fit on screen.) xxx

    mainScreenFrame = [[NSScreen mainScreen] frame];

    miny = - (dockTileOrigin.y - mainScreenFrame.origin.y) / STANDARD_TILE_HEIGHT;
    maxy = (NSMaxY (mainScreenFrame) - (dockTileOrigin.y + STANDARD_TILE_HEIGHT)) / STANDARD_TILE_HEIGHT;

    minx = - (dockTileOrigin.x - mainScreenFrame.origin.x) / STANDARD_TILE_WIDTH;
    maxx = (NSMaxX (mainScreenFrame) - (dockTileOrigin.x + STANDARD_TILE_WIDTH)) / STANDARD_TILE_WIDTH;


    defaults = [NSUserDefaults standardUserDefaults];
    fiendTileCorner = [defaults objectForKey:DK_FiendTileCorner];
    placeTilesVertically = [defaults boolForKey:DK_PlaceTilesVertically];

    // Two cases, place either vertically or horizontally first.
    // Then, depending on the corner, we set the start x,y, end x,y, and delta x,y's.
    // The start x,y are always 0,0.

    if ([fiendTileCorner isEqual:@"SW"] == YES)
    {
        endX = maxx;
        endY = maxy;
        deltaX = 1;
        deltaY = 1;
    }
    else if ([fiendTileCorner isEqual:@"NW"] == YES)
    {
        endX = maxx;
        endY = miny;
        deltaX = 1;
        deltaY = -1;
    }
    else if ([fiendTileCorner isEqual:@"SE"] == YES)
    {
        endX = minx;
        endY = maxy;
        deltaX = -1;
        deltaY = 1;
    }
    else // NE
    {
        endX = minx;
        endY = miny;
        deltaX = -1;
        deltaY = -1;
    }

    //NSLog (@"end: (%d,%d), delta: (%d,%d)", endX, endY, deltaX, deltaY);

    // Scan persistent and current dockLevels for empty tile space

    if (placeTilesVertically == YES)
    {
        //NSLog (@"Placing tiles vertically.");
        for (current.x = 0; current.x != (endX + deltaX) ; current.x += deltaX)
        {
            for (current.y = 0; current.y != (endY + deltaY); current.y += deltaY)
            {
                // The (0,0) location is reserved for the dock window.
                if (current.x == 0 && current.y == 0)
                    continue;
                if (current.x == aLocation.x && current.y == aLocation.y)
                {
                    //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);
                    return current;
                }
            
                if ([persistentLevel isDockLocationEmpty:current] == YES
                    && [currentLevel isDockLocationEmpty:current] == YES)
                {
                    //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);
                    return current;
                }
            }
        }
    }
    else
    {
        //NSLog (@"Placing tiles horizontally.");
        for (current.y = 0; current.y != (endY + deltaY); current.y += deltaY)
        {
            for (current.x = 0; current.x != (endX + deltaX) ; current.x += deltaX)
            {
                // The (0,0) location is reserved for the dock window.
                if (current.x == 0 && current.y == 0)
                    continue;
                if (current.x == aLocation.x && current.y == aLocation.y)
                {
                    //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);
                    return current;
                }
            
                if ([persistentLevel isDockLocationEmpty:current] == YES
                    && [currentLevel isDockLocationEmpty:current] == YES)
                {
                    //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);
                    return current;
                }
            }
        }
    }

    NSLog (@"No dock locations available.");

    //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);

    // We never want to cover up the main Fiend tile. (Perhaps it should have a higher window level...)
    current.x = deltaX;
    current.y = deltaY;

    return current;
}

//----------------------------------------------------------------------

// This was the initial placement algorithm
- (DockLocation) nextAvailableDockLocation2:(NSPoint)dockTileOrigin orLocation:(DockLocation)aLocation
{
    int minx, miny, maxx, maxy;
    DockLocation current;
    NSRect mainScreenFrame;

    // Calculate number of tiles the will fit below, left, above, and right of the main tile.
    // (Calculate total number of tiles that will fit on screen.) xxx

    mainScreenFrame = [[NSScreen mainScreen] frame];

    miny = - (dockTileOrigin.y - mainScreenFrame.origin.y) / STANDARD_TILE_HEIGHT;
    maxy = (NSMaxY (mainScreenFrame) - (dockTileOrigin.y + STANDARD_TILE_HEIGHT)) / STANDARD_TILE_HEIGHT;

    minx = - (dockTileOrigin.x - mainScreenFrame.origin.x) / STANDARD_TILE_WIDTH;
    maxx = (NSMaxX (mainScreenFrame) - (dockTileOrigin.x + STANDARD_TILE_WIDTH)) / STANDARD_TILE_WIDTH;

    //SNLogRect (mainScreenFrame);
    //SNLogRect (originFrame);
    //NSLog (@"miny: %d, minx: %d, maxy: %d, maxx: %d", miny, minx, maxy, maxx);

    // Scan persistent and current dockLevels for empty tile space

    for (current.x = 0; current.x >= minx; current.x--)
    {
        for (current.y = 0; current.y >= miny; current.y--)
        {
            // The (0,0) location is reserved for the dock window.
            if (current.x == 0 && current.y == 0)
                continue;
            if (current.x == aLocation.x && current.y == aLocation.y)
                return current;
            
            if ([persistentLevel isDockLocationEmpty:current] == YES
                && [currentLevel isDockLocationEmpty:current] == YES)
            {
                //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);
                return current;
            }
        }
    }

    for (current.x = 1; current.x <= maxx; current.x++)
    {
        for (current.y = 0; current.y >= miny; current.y--)
        {
            if (current.x == aLocation.x && current.y == aLocation.y)
                return current;

            if ([persistentLevel isDockLocationEmpty:current] == YES
                && [currentLevel isDockLocationEmpty:current] == YES)
            {
                //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);
                return current;
            }
        }
    }

    for (current.x = 0; current.x >= minx; current.x--)
    {
        for (current.y = 1; current.y <= maxy; current.y++)
        {
            if (current.x == aLocation.x && current.y == aLocation.y)
                return current;

            if ([persistentLevel isDockLocationEmpty:current] == YES
                && [currentLevel isDockLocationEmpty:current] == YES)
            {
                //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);
                return current;
            }
        }
    }

    for (current.x = 1; current.x <= maxx; current.x++)
    {
        for (current.y = 1; current.y <= maxy; current.y++)
        {
            if (current.x == aLocation.x && current.y == aLocation.y)
                return current;

            if ([persistentLevel isDockLocationEmpty:current] == YES
                && [currentLevel isDockLocationEmpty:current] == YES)
            {
                //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);
                return current;
            }
        }
    }

    NSLog (@"No dock locations available.");

    //NSLog (@"Dock Location: (%d,%d)", current.x, current.y);

    current.x = 0;
    current.y = 0;

    return current;
}
    
//----------------------------------------------------------------------

// This only checks the persistent and current levels, but we generally
// only want to know about the visible locations.
- (BOOL) isLocationAvailable:(DockLocation)aLocation
{
    NSEnumerator *tileEnumerator;
    FITile *tile;
    BOOL available;
    DockLocation tileLocation;

    // (0,0) is always reserved.
    if (aLocation.x == 0 && aLocation.y == 0)
        return NO;

    available = YES;
    tileEnumerator = [[persistentLevel tiles] objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        tileLocation = [tile location];
        if (tileLocation.x == aLocation.x && tileLocation.y == aLocation.y)
        {
            available = NO;
            break;
        }
    }

    if (available == YES)
    {
        tileEnumerator = [[currentLevel tiles] objectEnumerator];
        while (tile = [tileEnumerator nextObject])
        {
            tileLocation = [tile location];
            if (tileLocation.x == aLocation.x && tileLocation.y == aLocation.y)
            {
                available = NO;
                break;
            }
        }
    }

    return available;
}

//----------------------------------------------------------------------

- (BOOL) containsApplication:(NSString *)applicationPath;
{
    NSEnumerator *levelEnumerator;
    FILevel *level;
    BOOL isDocked;

    isDocked = [persistentLevel containsApplication:applicationPath];
    if (isDocked == NO)
    {
        levelEnumerator = [levels objectEnumerator];
        while (level = [levelEnumerator nextObject])
        {
            isDocked = [level containsApplication:applicationPath];
            if (isDocked == YES)
                break;
        }
    }

    return isDocked;
}

//----------------------------------------------------------------------
// For the GoToLevel panel
//----------------------------------------------------------------------

// Make a list of the dock level identifiers.  The identifier is either
// the dock level name, or N-M, where N is the dock level, and M is the
// total number of levels.

- (NSArray *) levelIdentifiers
{
    NSMutableArray *identifiers;
    NSEnumerator *levelEnumerator;
    FILevel *level;
    int index, levelCount;
    NSString *levelName;

    identifiers = [NSMutableArray array];

    index = 1;
    levelCount = [levels count];
    levelEnumerator = [levels objectEnumerator];
    while (level = [levelEnumerator nextObject])
    {
        levelName = [level levelName];
        if (levelName == nil)
        {
            levelName = [NSString stringWithFormat:@"%d-%d", index, levelCount];
        }

        [identifiers addObject:levelName];

        index++;
    }

    return identifiers;
}

//----------------------------------------------------------------------

- (int) indexOfCurrentLevel
{
    return [levels indexOfObject:currentLevel];
}

//----------------------------------------------------------------------

// Or would this be better as (goToDockLevelNumber:(int)aNumber?

- (void) goToLevelAtIndex:(int)index
{
    [self setCurrentLevel:[levels objectAtIndex:index]];
}

//----------------------------------------------------------------------
// NSWorkspace notifications
//----------------------------------------------------------------------

- (void) applicationWillLaunch:(NSNotification *)aNotification
{
    [persistentLevel applicationWillLaunch:aNotification];
    [levels makeObjectsPerformSelector:@selector (applicationWillLaunch:) withObject:aNotification];
}

//----------------------------------------------------------------------

- (void) applicationDidLaunch:(NSNotification *)aNotification
{
    [persistentLevel applicationDidLaunch:aNotification];
    [levels makeObjectsPerformSelector:@selector (applicationDidLaunch:) withObject:aNotification];
}

//----------------------------------------------------------------------

- (void) applicationDidTerminate:(NSNotification *)aNotification
{
    [persistentLevel applicationDidTerminate:aNotification];
    [levels makeObjectsPerformSelector:@selector (applicationDidTerminate:) withObject:aNotification];
}

//----------------------------------------------------------------------
// Fiend Launcher notifications
//----------------------------------------------------------------------

- (void) applicationLaunchFailed:(NSNotification *)aNotification
{
    NSString *path;

    path = [[aNotification userInfo] objectForKey:FILauncherApplicationPath];
    //SNLog (path);

    [persistentLevel applicationLaunchFailed:aNotification];
    [levels makeObjectsPerformSelector:@selector (applicationLaunchFailed:) withObject:aNotification];
}

//----------------------------------------------------------------------
// Preference notifications
//----------------------------------------------------------------------

- (void) hideTilesDidChange:(NSNotification *)aNotification
{
    if ([[NSUserDefaults standardUserDefaults] boolForKey:DK_HideTiles] == YES)
        [self hideSourceTiles];
    else
        [self unhideSourceTiles];
}

//----------------------------------------------------------------------

- (void) showLevelNumberDidChange:(NSNotification *)aNotification
{
    [self updateDockView];
}

//----------------------------------------------------------------------

- (void) showFlagCuesDidChange:(NSNotification *)aNotification
{
    [self updateTileViews];
}

//----------------------------------------------------------------------
// Other
//----------------------------------------------------------------------

- (void) unhide
{
    [persistentLevel setTileWindowLevel:[self tileWindowLevel]];
    [persistentLevel unhideLevel];

    [currentLevel setTileWindowLevel:[self tileWindowLevel]];
    [currentLevel unhideLevel];
}

//----------------------------------------------------------------------

- (void) hide
{
    [persistentLevel hideLevel];
    [currentLevel hideLevel];
}

//----------------------------------------------------------------------

- (void) checkForActiveApplications:(NSDictionary *)activeApplications
{
    [persistentLevel checkForActiveApplications:activeApplications];
    [levels makeObjectsPerformSelector:@selector (checkForActiveApplications:) withObject:activeApplications];
}

//----------------------------------------------------------------------

- (void) performAutoLaunch:sender
{
    [persistentLevel performAutoLaunch];
    [levels makeObjectsPerformSelector:@selector (performAutoLaunch)];
}

//----------------------------------------------------------------------

- (void) hideSourceTiles
{
    [persistentLevel hideSourceTiles];
    [levels makeObjectsPerformSelector:@selector (hideSourceTiles)];
}

//----------------------------------------------------------------------

- (void) unhideSourceTiles
{
    [persistentLevel unhideSourceTiles];
    [levels makeObjectsPerformSelector:@selector (unhideSourceTiles)];
}

//----------------------------------------------------------------------

- (void) updateTileViews
{
    [persistentLevel updateTileViews];
    [levels makeObjectsPerformSelector:@selector (updateTileViews)];
}

//----------------------------------------------------------------------

- (void) updateChangeCount
{
    if (nonretainedDelegate != nil && [nonretainedDelegate respondsToSelector:@selector (updateChangeCount)] == YES)
    {
        [nonretainedDelegate updateChangeCount];
    }
}

//----------------------------------------------------------------------

- (void) updateDockView
{
    if (nonretainedDelegate != nil && [nonretainedDelegate respondsToSelector:@selector (updateDockView)] == YES)
    {
        [nonretainedDelegate updateDockView];
    }
}

//----------------------------------------------------------------------

- (void) levelsDidChange
{
    [[NSNotificationCenter defaultCenter] postNotificationName:FIDockLevelsDidChangeNotification object:self];
}

//----------------------------------------------------------------------

- (void) repositionTilesRelativeToPoint:(NSPoint)newOrigin
{
    NSEnumerator *levelEnumerator;
    FILevel *level;

    [persistentLevel repositionRelativeToPoint:newOrigin];
    levelEnumerator = [levels objectEnumerator];
    while (level = [levelEnumerator nextObject])
    {
        [level repositionRelativeToPoint:newOrigin];
    }
}

//----------------------------------------------------------------------

- (void) repositionVisibleTilesRelativeToPoint:(NSPoint)newOrigin
{
    [persistentLevel repositionRelativeToPoint:newOrigin];
    [currentLevel repositionRelativeToPoint:newOrigin];
}

//----------------------------------------------------------------------

- (void) togglePersistenceOfTile:(FITile *)aTile
{
    FILevel *level;
    
    [aTile retain];

    // Doesn't do anything about preserving location.

    if ([aTile isPersistent] == YES)
    {
        [persistentLevel removeTile:aTile];
        [currentLevel addTile:aTile];
        [aTile setIsPersistent:NO];
    }
    else
    {
        level = [aTile level];
        [level removeTile:aTile];
        [persistentLevel addTile:aTile];
        [aTile setIsPersistent:YES];
    }

    [self updateChangeCount];

    [aTile release];
}

@end
