//
// $Id: FIDockView.h,v 1.1.1.1 1998/09/03 01:14:52 nygard Exp $
//

#import <AppKit/AppKit.h>

@protocol FIDockViewDatasourceProtocol
- (NSString *) currentLevelNumber;
- (NSString *) currentLevelName;
- (BOOL) isVisible;
- (BOOL) isLocked;
@end

@protocol FIDockViewDelegateProtocol
- (void) addApplicationsToCurrentLevel:(NSArray *)applications;
- (void) dockWindowDidDrag:(NSPoint)newWindowOrigin;
@end

//
// Original Fiend:
//
// ! mouse-up in left eye does a "twist"/hahaha
// ! mouse-up in right eye does a "fade"/yell
// - mouse-up toggles level visibility
// x shift-alt up toggles non-persistent level visibility (Change this to just, plain, shift-up.)
// * command-up goes to next level (There are little buttons there already, so USE them!)
// * shift-command-up goes to previous level (There are little buttons there already, so USE them!)
// x shift-up toggles lock (Change this to control-up.)
// - alt-up toggles raise/sink
// * control-up rotates clockwise (Use, say, command-click next/previous level buttons for clockwise/counter rotation.)
// - dragging non-locked dock
// ? shift-dragging can move locked dock (Change this to control-dragging?)
// ? shift-dragging while tiles selected will move those tiles off of dock (Change this to control-dragging?)
//
// Note: you can hide non-persistent level, but when you show the level, it is always shown.
//

@interface FIDockView : NSView
{
    NSButton *previousLevelButton;
    NSButton *activateButton;
    NSButton *nextLevelButton;

    NSWindow *dragWindow;
    NSImageView *dragWindowImageView;
    NSImage *dragWindowImage;

    id nonretainedDatasource;
    id nonretainedDelegate;
}

+ (void) initialize;

+ (void) adjustDockCellFonts;

- initWithFrame:(NSRect)frameRect;
- (void) dealloc;

- (NSWindow *) dragWindow;
- (void) updateDragWindowImage;

- (void) drawRect:(NSRect)rect;

- (void) mouseDown:(NSEvent *)theEvent;

- datasource;
- (void) setDatasource:newDatasource;

- delegate;
- (void) setDelegate:newDelegate;

//----------------------------------------------------------------------
// Button actions
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// Covers for delegate methods
//----------------------------------------------------------------------

- (void) addApplicationsToCurrentLevel:(NSArray *)applications;

// Sent to first responder
- (void) nextLevel:sender;
- (void) previousLevel:sender;
- (void) activateApplication:sender;
- (void) toggleDockVisible;
- (void) toggleDockSunken;
- (void) toggleDockLocked;

//----------------------------------------------------------------------
// Covers for datasource methods
//----------------------------------------------------------------------

- (BOOL) isLocked;

//----------------------------------------------------------------------
// NSDraggingDestination
//----------------------------------------------------------------------

- (BOOL) prepareForDragOperation:(id <NSDraggingInfo>)sender;
- (BOOL) performDragOperation:(id <NSDraggingInfo>)sender;
- (void) concludeDragOperation:(id <NSDraggingInfo>)sender;

- (unsigned int) draggingEntered:(id <NSDraggingInfo>)sender;
- (unsigned int) draggingUpdated:(id <NSDraggingInfo>)sender;
- (void) draggingExited:(id <NSDraggingInfo>)sender;

@end
