#import "Fiend.h"

RCSID ("$Id: FIDockView.m,v 1.2 1998/09/03 01:22:37 nygard Exp $");

#import "FIDockView.h"

#import "FIDock.h"
#import "NSImage+SNExtensions.h"
#import "Fiendish.h"
#import "FISoundManager.h"

#include <sys/time.h>

// Some of these class variables would need to be changed if it were to support
// more than one DockManagerView per application.  It may be better to anyways.

static NSImage *_blankTileImage = nil;
static NSImage *_fiendImage = nil;
static NSImage *_alternateFiendImage = nil;

static NSImage *_leftTriangleImage = nil;
static NSImage *_leftTriangleHImage = nil;
static NSImage *_diamondImage = nil;
static NSImage *_diamondHImage = nil;
static NSImage *_rightTriangleImage = nil;
static NSImage *_rightTriangleHImage = nil;

static NSTextFieldCell *_dockNameCell = nil;
static NSTextFieldCell *_dockNumberCell = nil;

#define TILE_WIDTH (64)
#define TILE_HEIGHT (64)

@interface NSApplication (hidden)
- (int) _lastEventTime;
@end


@implementation FIDockView

+ (void) initialize
{
    NSBundle *thisBundle;

    if (self == [FIDockView class])
    {
        thisBundle = [NSBundle bundleForClass:self];
        NSAssert (thisBundle != nil, @"Could not get bundle.");

        _blankTileImage = [[NSImage imageNamed:@"BlankTile.tiff" fromBundle:thisBundle] retain];
        _fiendImage = [[NSImage imageNamed:@"FiendIcon.tiff" fromBundle:thisBundle] retain];
        _alternateFiendImage = [[NSImage imageNamed:@"AltFiendIcon.tiff" fromBundle:thisBundle] retain];

        _leftTriangleImage = [[NSImage imageNamed:@"leftTriangle.tiff" fromBundle:thisBundle] retain];
        _leftTriangleHImage = [[NSImage imageNamed:@"leftTriangleH.tiff" fromBundle:thisBundle] retain];
        _diamondImage = [[NSImage imageNamed:@"diamond.tiff" fromBundle:thisBundle] retain];
        _diamondHImage = [[NSImage imageNamed:@"diamondH.tiff" fromBundle:thisBundle] retain];
        _rightTriangleImage = [[NSImage imageNamed:@"rightTriangle.tiff" fromBundle:thisBundle] retain];
        _rightTriangleHImage = [[NSImage imageNamed:@"rightTriangleH.tiff" fromBundle:thisBundle] retain];

        _dockNameCell = [[NSTextFieldCell allocWithZone:[self zone]] initTextCell:@"1-1"];
        [_dockNameCell setAlignment:NSLeftTextAlignment];
        [_dockNameCell setDrawsBackground:NO];
        [_dockNameCell setTextColor:[NSColor blackColor]];
        [_dockNameCell setEditable:YES];
        [_dockNameCell setEnabled:YES];
        [_dockNameCell setWraps:NO]; // For RDR2
        //[_dockNameCell setFont:[NSFont controlContentFontOfSize:10]];

        _dockNumberCell = [[NSTextFieldCell allocWithZone:[self zone]] initTextCell:@"1"];
        [_dockNumberCell setAlignment:NSLeftTextAlignment];
        [_dockNumberCell setDrawsBackground:NO];
        [_dockNumberCell setTextColor:[NSColor blackColor]];
        [_dockNumberCell setEditable:YES];
        [_dockNumberCell setEnabled:YES];
        //[_dockNumberCell setFont:[NSFont controlContentFontOfSize:10]];

        [self adjustDockCellFonts];
    }
}

//----------------------------------------------------------------------

+ (void) adjustDockCellFonts
{
    NSUserDefaults *defaults;
    float fontSize;
    BOOL useBoldDockFont;
    BOOL useBigDockFont;

    defaults = [NSUserDefaults standardUserDefaults];
    useBoldDockFont = [defaults boolForKey:DK_UseBoldDockFont];
    useBigDockFont = [defaults boolForKey:DK_UseBigDockFont];

    //NSLog (@"useBold: %@, useBig: %@", ((useBoldDockFont == YES) ? @"Yes" : @"No"), ((useBigDockFont == YES) ? @"Yes" : @"No"));

    fontSize = (useBigDockFont == YES) ? 12.0 : 10.0;
#if 0
    fontSize *= (tileWidth / STANDARD_TILE_WIDTH);
    fontSize = (fontSize < 6.0) ? 6.0 : fontSize;
    fontSize = (fontSize > 18.0) ? 18.0 : fontSize;
    fontSize = rint (fontSize);

    NSLog (@"fontSize: %f", fontSize);
#endif
#if 0
    if (useBoldDockFont == YES)
    {
        [_dockNameCell setFont:[NSFont boldSystemFontOfSize:fontSize]];
        [_dockNumberCell setFont:[NSFont boldSystemFontOfSize:fontSize - 1.0]];
    }
    else
    {
    }
#endif
    [_dockNameCell setFont:[NSFont controlContentFontOfSize:fontSize]];
    [_dockNumberCell setFont:[NSFont controlContentFontOfSize:fontSize - 1.0]];
}



//----------------------------------------------------------------------

- initWithFrame:(NSRect)frameRect
{
    if ([super initWithFrame:frameRect] == nil)
        return nil;

    dragWindow = nil;
    dragWindowImageView = nil;
    dragWindowImage = nil;

    [self registerForDraggedTypes:[NSArray arrayWithObject:NSFilenamesPboardType]];

    previousLevelButton = [[NSButton alloc] initWithFrame:NSMakeRect (51, 4, 9, 9)];
    [previousLevelButton setImage:_leftTriangleImage];
    [previousLevelButton setAlternateImage:_leftTriangleHImage];
    [previousLevelButton setBordered:NO];
    [previousLevelButton setImagePosition:NSImageOnly];
    [previousLevelButton setButtonType:NSMomentaryChangeButton];
    [previousLevelButton setEnabled:YES];

    [previousLevelButton setTarget:self];
    [previousLevelButton setAction:@selector (previousLevel:)];

    //[previousLevelButton setTransparent:YES];
    [self addSubview:previousLevelButton];

    activateButton = [[NSButton alloc] initWithFrame:NSMakeRect (51, 4+9, 9, 10)];
    [activateButton setImage:_diamondImage];
    [activateButton setAlternateImage:_diamondHImage];
    [activateButton setBordered:NO];
    [activateButton setImagePosition:NSImageOnly];
    [activateButton setButtonType:NSMomentaryChangeButton];
    [activateButton setEnabled:YES];

    [activateButton setTarget:self];
    [activateButton setAction:@selector (activateApplication:)];

    [self addSubview:activateButton];

    nextLevelButton = [[NSButton alloc] initWithFrame:NSMakeRect (51, 4+9+10, 9, 9)];
    [nextLevelButton setImage:_rightTriangleImage];
    [nextLevelButton setAlternateImage:_rightTriangleHImage];
    [nextLevelButton setBordered:NO];
    [nextLevelButton setImagePosition:NSImageOnly];
    [nextLevelButton setButtonType:NSMomentaryChangeButton];

    [nextLevelButton setTarget:self];
    [nextLevelButton setAction:@selector (nextLevel:)];

    [self addSubview:nextLevelButton];

    //NSLog (@"target: %@", [previousLevelButton target]);

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    SNRelease (dragWindow);
    SNRelease (dragWindowImageView);
    SNRelease (dragWindowImage);
    
    [super dealloc];
}

//----------------------------------------------------------------------

- (NSWindow *) dragWindow
{
    NSWindow *dockWindow;
    NSRect windowFrame, imageViewFrame;
    
    if (dragWindow == nil)
    {
        NSAssert (dragWindowImage == nil, @"Expected dragWindowImage to be nil when dragWindow is nil.");

        dockWindow = [self window];
        windowFrame = [dockWindow frame];

        dragWindow = [[NSWindow alloc] initWithContentRect:windowFrame
                                       styleMask:NSBorderlessWindowMask
                                       backing:NSBackingStoreBuffered
                                       defer:NO];
        [dragWindow setLevel:[dockWindow level]];

        imageViewFrame.origin = NSZeroPoint;
        imageViewFrame.size = windowFrame.size;
        
        dragWindowImage = [[NSImage alloc] initWithSize:imageViewFrame.size];
        //[self updateDragWindowImage];

        dragWindowImageView = [[NSImageView alloc] initWithFrame:imageViewFrame];
        [dragWindowImageView setImageAlignment:NSImageAlignBottomLeft];
        [dragWindowImageView setImageFrameStyle:NSImageFrameNone];
        [dragWindowImageView setImageScaling:NSScaleNone];
        [dragWindowImageView setImage:dragWindowImage];
        //[image release];

        [dragWindow setContentView:dragWindowImageView];
        //[imageView release];
    }

    return dragWindow;
}

//----------------------------------------------------------------------

- (void) updateDragWindowImage
{
    NSWindow *window;
    int sourceWindowNumber;
    NSRect windowFrame;
    int status;

    window = [self window];
    windowFrame = [window frame];
    windowFrame.origin = NSZeroPoint;
    NSConvertWindowNumberToGlobal ([window windowNumber], &sourceWindowNumber);

    [dragWindowImage lockFocus];
    status = safeCopyWindow (windowFrame.origin.x, windowFrame.origin.y, windowFrame.size.width, windowFrame.size.height,
                             sourceWindowNumber);
    [dragWindowImage unlockFocus];
    if (status != 0)
    {
        NSLog (@"safeCopyWindow() failed.");
    }

    // Note: We can't just recache the image, since the view appears to copy it.
    //[dragWindowImage recache];
    [dragWindowImageView setImage:dragWindowImage];
}

//----------------------------------------------------------------------

- (void) drawRect:(NSRect)rect
{
    NSPoint aPoint;
    NSSize imageSize;
    NSSize dockNameCellSize;
    NSRect dockNumberFrame;
    NSRect dockNameFrame;
    BOOL isVisible;
    NSImage *tileImage;
    NSUserDefaults *defaults;

    defaults = [NSUserDefaults standardUserDefaults];
    
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (currentLevelNumber)] == YES)
    {
        [_dockNumberCell setStringValue:[nonretainedDatasource currentLevelNumber]];
    }

    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (currentLevelName)] == YES)
    {
        [_dockNameCell setStringValue:[nonretainedDatasource currentLevelName]];
    }

    if (_blankTileImage != nil)
    {
        [_blankTileImage compositeToPoint:NSZeroPoint operation:NSCompositeSourceOver];
    }

    isVisible = YES;
#if 1
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (isVisible)] == YES)
    {
        isVisible = [nonretainedDatasource isVisible];
    }
#endif
    if (_fiendImage != nil)
    {
        tileImage = (isVisible == YES) ? _fiendImage : _alternateFiendImage;
        imageSize = [tileImage size];
        dockNameCellSize = [_dockNameCell cellSize];
        aPoint.x = rint ((TILE_WIDTH - 6 - imageSize.width) / 2.0);
        aPoint.y = rint ((TILE_HEIGHT - dockNameCellSize.height - imageSize.height) / 2.0);
#if 0
        [[NSColor blueColor] set];
        NSRectFill (NSMakeRect (aPoint.x, aPoint.y, imageSize.width, imageSize.height));
#endif
        [tileImage compositeToPoint:aPoint operation:NSCompositeSourceOver];
    }

    dockNameFrame.size = [_dockNameCell cellSize];
    dockNameFrame.origin.x = 2;
    // Skip over the top pixel border (should be 2 pixels...)
    dockNameFrame.origin.y = TILE_HEIGHT - 1 - dockNameFrame.size.height;
    [_dockNameCell drawWithFrame:dockNameFrame inView:self];

    if ([defaults boolForKey:DK_ShowLevelNumber] == YES)
    {
        // Skip over the 2 pixel tile border
        dockNumberFrame.origin = NSMakePoint (2, 2); 
        dockNumberFrame.size = [_dockNumberCell cellSize];
        [_dockNumberCell drawWithFrame:dockNumberFrame inView:self];
    }

    if ([self isLocked] == YES)
    {
        [[NSColor redColor] set];
        PSarc (55, 38, 1.8, 0, 360);
        PSfill ();
        PSsetgray (NSBlack);
        PSsetlinewidth (0);
        PSarc (55, 38, 1.8, 0, 360);
        PSstroke ();
    }
}

//----------------------------------------------------------------------

//
// It just doesn't work to move the window while dragging from it, if you
// want to get meaningful mouse locations -- the mouse locations are in the
// window coordinates, and as soon as you move it, *poof*, the locations
// of any events generated up to the time you moved it are meaningless.
// And, of course, there seems to be no way whatsoever to figure out when
// in the event stream you move the window. Arrgh!
//
// New idea -- don't move the window.  Leave it there, but hide it.  Then,
// snap an image of the tile that you will put in a window that you *can* move,
// and move that, with the tiles tracking it.
//
// Even better, just get the events from another window that *isn't* moving,
// and all should be well.  I haven't tried this yet.
//

- (void) mouseDown:(NSEvent *)theEvent
{
    NSPoint originalWindowOrigin, originalScreenPoint;
    NSWindow *window;

    NSPoint mouseLocation;
    NSPoint newWindowOrigin;

    unsigned int modifierFlags;

    modifierFlags = [theEvent modifierFlags];

    if ([theEvent clickCount] == 1)
    {
        window = [self window];
        originalWindowOrigin = [window frame].origin;
        originalScreenPoint = [window convertBaseToScreen:[theEvent locationInWindow]];

        //SNLogPoint (originalWindowOrigin);
        //SNLogPoint (originalScreenPoint);

        theEvent = [window nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
        if ([theEvent type] == NSLeftMouseDragged)
        {
            // You can only move the dock if it is not locked, or if you are control-dragging.
            // Yes, the modifier flags from the mouse down event are okay to use here.
            if ([self isLocked] == NO || (modifierFlags & NSControlKeyMask))
            {
                //----------
                [self dragWindow];
                [self updateDragWindowImage];
                [dragWindow orderFront:self];
                //----------
                [window orderOut:self];

                while ([theEvent type] != NSLeftMouseUp)
                {
                    if ([theEvent type] == NSLeftMouseDragged)
                    {
                        mouseLocation = [window convertBaseToScreen:[theEvent locationInWindow]];
                        //SNLogPoint (mouseLocation);
                        newWindowOrigin.x = originalWindowOrigin.x + mouseLocation.x - originalScreenPoint.x;
                        newWindowOrigin.y = originalWindowOrigin.y + mouseLocation.y - originalScreenPoint.y;

                        [dragWindow setFrameOrigin:newWindowOrigin];
                        [nonretainedDelegate dockWindowDidDrag:newWindowOrigin];
                    }

                    theEvent = [window nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
                }

                [window setFrameOrigin:[dragWindow frame].origin];
                [window orderFront:self];
                [dragWindow orderOut:self];
            }
            else
            {
                NSBeep ();

                // And gobble events up to and including the mouse up.
                theEvent = [window nextEventMatchingMask:NSLeftMouseUpMask];
                [window discardEventsMatchingMask:NSLeftMouseDraggedMask beforeEvent:theEvent];
            }
        }
        else
        {
            // If we actually dragged, gobble up the mouse up event, otherwise pass it on.
            [self mouseUp:theEvent];
        }
    }
    else if ([theEvent clickCount] == 2 && (modifierFlags & NSCommandKeyMask))
    {
        [[NSWorkspace sharedWorkspace] hideOtherApplications];
    }
}

//----------------------------------------------------------------------

- (void) mouseUp:(NSEvent *)theEvent
{
    unsigned int modifierFlags;
    int control, command, alternate, shift;

    modifierFlags = [theEvent modifierFlags];
    control = modifierFlags & NSControlKeyMask;
    command = modifierFlags & NSCommandKeyMask;
    alternate = modifierFlags & NSAlternateKeyMask;
    shift = modifierFlags & NSShiftKeyMask;
    
    //NSLog (@"modifierFlags: %08x", modifierFlags);
    if (modifierFlags & NSAlternateKeyMask)
    {
        [self toggleDockSunken];
    }
    else if (modifierFlags & NSControlKeyMask)
    {
        [self toggleDockLocked];
    }
    else if (!control & !command & !alternate & !shift)
    {
        [self toggleDockVisible];
    }
}

//----------------------------------------------------------------------

- datasource
{
    return nonretainedDatasource;
}

//----------------------------------------------------------------------

- (void) setDatasource:newDatasource
{
    if (newDatasource == nonretainedDatasource)
        return;

    nonretainedDatasource = newDatasource;
}

//----------------------------------------------------------------------

- delegate
{
    return nonretainedDelegate;
}

//----------------------------------------------------------------------

- (void) setDelegate:newDelegate
{
    if (newDelegate == nonretainedDelegate)
        return;

    nonretainedDelegate = newDelegate;
}

//----------------------------------------------------------------------
// Button actions
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// Covers for delegate methods
//----------------------------------------------------------------------

- (void) nextLevel:sender
{
    [NSApp sendAction:@selector (nextLevel:) to:nil from:self];
}

//----------------------------------------------------------------------

- (void) previousLevel:sender
{
    [NSApp sendAction:@selector (previousLevel:) to:nil from:self];
}

//----------------------------------------------------------------------

- (void) activateApplication:sender
{
    [NSApp sendAction:@selector (activateApplication:) to:nil from:self];
}

//----------------------------------------------------------------------

- (void) addApplicationsToCurrentLevel:(NSArray *)applications
{
    if (nonretainedDelegate != nil && [nonretainedDelegate respondsToSelector:@selector (addApplicationsToCurrentLevel:)])
    {
        [nonretainedDelegate addApplicationsToCurrentLevel:applications];
    }
}

//----------------------------------------------------------------------

- (void) toggleDockVisible
{
    [NSApp sendAction:@selector (toggleDockVisible:) to:nil from:self];
}

//----------------------------------------------------------------------

- (void) toggleDockSunken
{
    [NSApp sendAction:@selector (toggleDockSunken:) to:nil from:self];
}

//----------------------------------------------------------------------

- (void) toggleDockLocked
{
    [NSApp sendAction:@selector (toggleDockLocked:) to:nil from:self];
}

//----------------------------------------------------------------------
// Covers for datasource methods
//----------------------------------------------------------------------

- (BOOL) isLocked
{
    BOOL flag;

    flag = YES;
    
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (isLocked)] == YES)
    {
        flag = [nonretainedDatasource isLocked];
    }

    return flag;
}

//----------------------------------------------------------------------
// NSDraggingDestination
//----------------------------------------------------------------------

- (BOOL) prepareForDragOperation:(id <NSDraggingInfo>)sender
{
    // App in dock, but not being dragged from out dock: reject
    // App in dock, but being dragged from out dock: remove it from dock

    // Reject a multiple selection if any of the applications are already on the dock
    return YES;
}

//----------------------------------------------------------------------

- (BOOL) performDragOperation:(id <NSDraggingInfo>)sender
{
    return YES;
}

//----------------------------------------------------------------------

- (void) concludeDragOperation:(id <NSDraggingInfo>)sender
{
    NSPasteboard *draggingPasteboard;
    NSArray *filenames;
    
    draggingPasteboard = [sender draggingPasteboard];
    //NSLog (@"draggingPasteboard: %@", draggingPasteboard);

    //NSLog (@"available types: %@", [draggingPasteboard types]);

    if ([draggingPasteboard availableTypeFromArray:[NSArray arrayWithObject:NSFilenamesPboardType]] != nil)
    {
        filenames = [draggingPasteboard propertyListForType:NSFilenamesPboardType];
        //NSLog (@"[%@] filenames: %@", NSStringFromClass ([filenames class]), filenames);

        [self addApplicationsToCurrentLevel:filenames];
        [[FISoundManager sharedSoundManager] didAddTiles];
    }
}

//----------------------------------------------------------------------

- (unsigned int) draggingEntered:(id <NSDraggingInfo>)sender
{
    return NSDragOperationCopy;
}

//----------------------------------------------------------------------

- (unsigned int) draggingUpdated:(id <NSDraggingInfo>)sender
{
    return NSDragOperationCopy;
}

//----------------------------------------------------------------------

- (void) draggingExited:(id <NSDraggingInfo>)sender
{
}

@end
