#import "Fiend.h"

RCSID ("$Id: FILauncher.m,v 1.1.1.1 1998/09/03 01:14:56 nygard Exp $");

#import "FILauncher.h"

DEFINE_NSSTRING (FIFailedToLaunchApplicationNotificaiton);
DEFINE_NSSTRING (FILauncherApplicationPath);

@implementation FILauncher

+ sharedLauncher
{
    static FILauncher *instance = nil;

    if (instance == nil)
        instance = [[FILauncher alloc] init];

    return instance;
}

//----------------------------------------------------------------------

- init
{
    NSBundle *thisBundle;
    NSString *fiendLauncherPath;
    
    if ([super init] == nil)
        return nil;

    thisBundle = [NSBundle bundleForClass:[self class]];
    NSAssert (thisBundle != nil, @"Couldn't get bundle.");

    fiendLauncherPath = [thisBundle pathForResource:@"FiendLauncher" ofType:nil];

    launcherTask = [[NSTask alloc] init];
    [launcherTask setLaunchPath:fiendLauncherPath];

    writePipe = [[NSPipe pipe] retain];
    writeHandle = [[writePipe fileHandleForWriting] retain];

    readPipe = [[NSPipe pipe] retain];
    readHandle = [[readPipe fileHandleForReading] retain];
    readBuffer = [[NSMutableData data] retain];

    [[NSNotificationCenter defaultCenter] addObserver:self
                                          selector:@selector (readData:)
                                          //name:NSFileHandleReadCompletionNotification
                                          name:NSFileHandleDataAvailableNotification
                                          object:readHandle];
    [readHandle waitForDataInBackgroundAndNotify];

    [launcherTask setStandardInput:writePipe];
    [launcherTask setStandardOutput:readPipe];

    NS_DURING
        {
            [launcherTask launch];
        }
    NS_HANDLER
        {
            EHAND;
        }
    NS_ENDHANDLER;

    lastLaunchPath = nil;

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];

    SNRelease (writeHandle);
    SNRelease (writePipe);
    SNRelease (readHandle);
    SNRelease (readBuffer);
    SNRelease (readPipe);
    SNRelease (launcherTask);

    [super dealloc];
}

//----------------------------------------------------------------------

- (void) launchApplication:(NSString *)path autolaunch:(BOOL)autolaunchFlag
{
    NSAssert ([launcherTask isRunning] == YES, @"Launcher task is not running.");

    [writeHandle writeString:[NSString stringWithFormat:@"%d %@\n", autolaunchFlag, path]];
    [self setLastLaunchPath:path];
}

//----------------------------------------------------------------------

- (NSString *) lastLaunchPath
{
    return lastLaunchPath;
}

//----------------------------------------------------------------------

- (void) setLastLaunchPath:(NSString *)newPath
{
    if (newPath == lastLaunchPath)
        return;

    SNRelease (lastLaunchPath);
    lastLaunchPath = [newPath retain];
}

//----------------------------------------------------------------------

- (NSString *) readLine
{
    NSString *readBufferString;
    NSRange newlineRange;
    NSString *str;
    
    readBufferString = [[[NSString alloc] initWithData:readBuffer encoding:NSASCIIStringEncoding] autorelease];
    newlineRange = [readBufferString rangeOfString:@"\n"];

    //NSLog (@"location: %d, length: %d", newlineRange.location, newlineRange.length);
    str = nil;
    if (newlineRange.length != 0)
    {
        NSData *oldBuffer;

        str = [readBufferString substringToIndex:newlineRange.location];

        oldBuffer = readBuffer;
        readBuffer = [[oldBuffer subdataWithRange:NSMakeRange (newlineRange.location + 1,
                                                               [oldBuffer length] - (newlineRange.location + 1))] mutableCopy];
        [oldBuffer release];
    }

    return str;
}

//----------------------------------------------------------------------

- (void) readData:(NSNotification *)aNotification
{
    NSString *line;
    
    [readBuffer appendData:[readHandle availableData]];
    [readHandle waitForDataInBackgroundAndNotify];

    while (line = [self readLine])
    {
        [[NSNotificationCenter defaultCenter] postNotificationName:FIFailedToLaunchApplicationNotificaiton
                                              object:self
                                              userInfo:[NSDictionary dictionaryWithObject:line forKey:FILauncherApplicationPath]];
        //SNLog (line);
    }
}

@end
