#import "Fiend.h"

RCSID ("$Id: FILevel.m,v 1.1.1.1 1998/09/03 01:14:51 nygard Exp $");

#import "FILevel.h"

#import "FITile.h"
#import "FIDock.h"

@implementation FILevel

+ (void) initialize
{
}

//----------------------------------------------------------------------

- initWithDock:(FIDock *)aDock
{
    if ([super init] == nil)
        return nil;

    nonretainedDock = aDock;

    levelName = nil;
    tiles = [[NSMutableArray array] retain];

    isPersistent = NO;

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    SNRelease (levelName);
    SNRelease (tiles);

    [super dealloc];
}

//----------------------------------------------------------------------

- (void) loadApplications:(NSArray *)savedApplications
{
    NSEnumerator *savedApplicationEnumerator;
    NSDictionary *savedApplication;
    FITile *newTile;
    
    savedApplicationEnumerator = [savedApplications objectEnumerator];
    while (savedApplication = [savedApplicationEnumerator nextObject])
    {
        newTile = [FITile tileFromPropertyList:savedApplication];
        //NSLog (@"newTile: %@", newTile);
        if (newTile != nil)
        {
            [newTile setIsPersistent:isPersistent];
            [self addTile:newTile];
        }
    }
}

//----------------------------------------------------------------------

- (NSDictionary *) levelPropertyList
{
    NSMutableDictionary *propertyList;
    NSMutableArray *tileList;
    NSEnumerator *tileEnumerator;
    FITile *tile;

    propertyList = [NSMutableDictionary dictionary];
    tileList = [NSMutableArray array];

    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        [tileList addObject:[tile tilePropertyList]];
    }

    if (levelName != nil)
    {
        [propertyList setObject:levelName forKey:@"DockLevelName"];
    }
    [propertyList setObject:tileList forKey:@"DockedApplications"];

    return propertyList;
}

//----------------------------------------------------------------------

- (FIDock *) dock
{
    return nonretainedDock;
}

//----------------------------------------------------------------------

- (NSString *) levelName
{
    return levelName;
}

//----------------------------------------------------------------------

- (void) setLevelName:(NSString *)newName
{
    if (newName == levelName)
        return;

    SNRelease (levelName);
    levelName = [newName copy];

    [nonretainedDock updateChangeCount];
    [nonretainedDock levelsDidChange];
}

//----------------------------------------------------------------------

- (BOOL) isPersistent
{
    return isPersistent;
}

//----------------------------------------------------------------------

- (void) setIsPersistent:(BOOL)flag
{
    isPersistent = flag;
}

//----------------------------------------------------------------------

- (NSArray *) tiles
{
    return tiles;
}

//----------------------------------------------------------------------

- (void) addTile:(FITile *)newTile
{
    [newTile setLevel:self];
    [tiles addObject:newTile];
}

//----------------------------------------------------------------------

- (BOOL) containsApplication:(NSString *)applicationPath
{
    NSEnumerator *tileEnumerator;
    FITile *tile;
    BOOL containsApplication;

    containsApplication = NO;

    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        if ([[tile path] isEqual:applicationPath] == YES)
        {
            containsApplication = YES;
            break;
        }
    }

    return containsApplication;
}

//----------------------------------------------------------------------

- (BOOL) isDockLocationEmpty:(DockLocation)aLocation
{
    NSEnumerator *tileEnumerator;
    FITile *tile;
    DockLocation currentLocation;
    BOOL isEmpty;

    isEmpty = YES;

    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        currentLocation = [tile location];
        if (currentLocation.x == aLocation.x && currentLocation.y == aLocation.y)
        {
            isEmpty = NO;
            break;
        }
    }

    return isEmpty;
}

//----------------------------------------------------------------------

- (void) hideLevel
{
    [tiles makeObjectsPerformSelector:@selector (hide)];
}

//----------------------------------------------------------------------

- (void) unhideLevel
{
    [tiles makeObjectsPerformSelector:@selector (unhide)];
}

//----------------------------------------------------------------------

- (void) setTileWindowLevel:(int)level
{
    NSEnumerator *tileEnumerator;
    FITile *tile;

    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        [[tile window] setLevel:level];
    }
}

//----------------------------------------------------------------------
// - Remove docked applications
// - return running dock tiles to original position?

- (void) clearLevel
{
    [tiles removeAllObjects];
}

//----------------------------------------------------------------------

- (void) rotateTilesAroundPoint:(NSPoint)origin
{
    NSEnumerator *tileEnumerator;
    FITile *tile;

    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        [tile rotateLocationClockwise:YES];
        [tile repositionRelativeToPoint:origin];
    }
}

//----------------------------------------------------------------------

- (void) repositionRelativeToPoint:(NSPoint)origin
{
    NSEnumerator *tileEnumerator;
    FITile *tile;
    DockLocation location;

    tileEnumerator = [tiles objectEnumerator];
    while (tile = [tileEnumerator nextObject])
    {
        location = [tile location];
        if (location.x != 0 || location.y != 0)
            [tile repositionRelativeToPoint:origin];
    }
}

//----------------------------------------------------------------------

- (void) checkForActiveApplications:(NSDictionary *)activeApplications
{
    [tiles makeObjectsPerformSelector:@selector (checkForActiveApplication:) withObject:activeApplications];
}

//----------------------------------------------------------------------

- (void) performAutoLaunch
{
    [tiles makeObjectsPerformSelector:@selector (autoLaunch)];
}

//----------------------------------------------------------------------

- (void) hideSourceTiles
{
    [tiles makeObjectsPerformSelector:@selector (hideSourceTile)];
}

//----------------------------------------------------------------------

- (void) unhideSourceTiles
{
    [tiles makeObjectsPerformSelector:@selector (unhideSourceTile)];
}

//----------------------------------------------------------------------

- (void) updateTileViews
{
    [tiles makeObjectsPerformSelector:@selector (updateTileView)];
}

//----------------------------------------------------------------------

- (void) deleteTiles:(NSArray *)someTiles
{
    [tiles removeObjectsInArray:someTiles];
}

//----------------------------------------------------------------------

// This shouldn't generate any noise... (i.e. don't play sound?)
- (void) removeTile:(FITile *)aTile
{
    [tiles removeObject:aTile];
}

//----------------------------------------------------------------------

// Pass along change notifications from tiles to dock...
- (void) updateChangeCount
{
    [nonretainedDock updateChangeCount];
}

//----------------------------------------------------------------------
// NSWorkspace notifications
//----------------------------------------------------------------------

- (void) applicationWillLaunch:(NSNotification *)aNotification
{
    [tiles makeObjectsPerformSelector:@selector (applicationWillLaunch:) withObject:aNotification];
}

//----------------------------------------------------------------------


- (void) applicationDidLaunch:(NSNotification *)aNotification
{
    [tiles makeObjectsPerformSelector:@selector (applicationDidLaunch:) withObject:aNotification];
}

//----------------------------------------------------------------------

- (void) applicationDidTerminate:(NSNotification *)aNotification
{
    [tiles makeObjectsPerformSelector:@selector (applicationDidTerminate:) withObject:aNotification];
}

//----------------------------------------------------------------------
// Fiend Launcher notifications
//----------------------------------------------------------------------

- (void) applicationLaunchFailed:(NSNotification *)aNotification
{
    [tiles makeObjectsPerformSelector:@selector (applicationLaunchFailed:) withObject:aNotification];
}

@end
