//
// $Id: FITile.h,v 1.1.1.1 1998/09/03 01:14:51 nygard Exp $
//

#import <AppKit/AppKit.h>

#import "Fiend.h"

@class FITileView;
@class FIAppInfo;
@class FILevel;

typedef enum _ApplicationState
{
    ApplicationInactiveState,
    ApplicationLaunching1State,
    ApplicationLaunching2State,
    ApplicationActiveState
} ApplicationState;

NSString *NSStringFromApplicationState (ApplicationState state);

@interface FITile : NSObject
{
    FILevel *nonretainedLevel; // XXX: So it can pass change messages back to the dock...
    
    NSString *path;
    DockLocation location;

    NSWindow *window;
    FITileView *iconView;

    // Options:
    BOOL shouldAutoLaunch;
    BOOL shouldAnimate;
    BOOL isPersistent;

    BOOL isSelected;

    ApplicationState state;
    NSTimer *appInfoTimer;
    FIAppInfo *appInfo;

    NSTimer *animationTimer;
    NSImage *animationImage;
}

+ (FITile *) tileFromPropertyList:(NSDictionary *)propertyList;
- (NSDictionary *) tilePropertyList;

- initWithRepresentedPath:(NSString *)aPath;
- (void) dealloc;

- (NSString *) path;
- (NSString *) shortName;
- (NSWindow *) window;

- (FILevel *) level;
- (void) setLevel:(FILevel *)newLevel;

- (BOOL) shouldAutoLaunch;
- (void) setShouldAutoLaunch:(BOOL)flag;

- (BOOL) shouldAnimate;
- (void) setShouldAnimate:(BOOL)flag;

- (BOOL) isPersistent;
- (void) setIsPersistent:(BOOL)flag;

- (BOOL) isRunning;

- (BOOL) isSelected;
- (void) setIsSelected:(BOOL)flag;

- (BOOL) isGhosted;

- (DockLocation) location;
- (void) setLocation:(DockLocation)newLocation;
- (void) rotateLocationClockwise:(BOOL)clockwise;

// Reserved for the dock window.
- (void) setWindow:(NSWindow *)newWindow;


//----------------------------------------------------------------------
// Actions...
//----------------------------------------------------------------------

- (void) createTileWindow;

- (void) repositionRelativeToPoint:(NSPoint)origin;

- (void) hide;
- (void) unhide;
- (void) setTileWindowLevel:(int)level;

- (void) checkForActiveApplication:(NSDictionary *)activeApplications;

//----------------------------------------------------------------------
// FITileView target methods
//----------------------------------------------------------------------

- (void) selectApplicationInWorkspace:sender;
- (void) autoLaunch;
- (void) hideSourceTile;
- (void) unhideSourceTile;
- (void) updateTileView;

- (void) launchApplication:sender hideOthers:(BOOL)hideOthers;
- (void) toggleTileSelected:sender;
//- (void) deselectAll:sender;

//----------------------------------------------------------------------
// NSWorkspace notifications
//----------------------------------------------------------------------

- (void) applicationWillLaunch:(NSNotification *)aNotification;
- (void) applicationDidLaunch:(NSNotification *)aNotification;
- (void) applicationDidTerminate:(NSNotification *)aNotification;

//----------------------------------------------------------------------
// Fiend Launcher notifications
//----------------------------------------------------------------------

- (void) applicationLaunchFailed:(NSNotification *)aNotification;

//----------------------------------------------------------------------
// App Info timer
//----------------------------------------------------------------------

- (void) startAppInfoTimer;
- (void) stopAppInfoTimer;
- (void) fireAppInfoTimer:(NSTimer *)aTimer;

//----------------------------------------------------------------------
// Animation timer
//----------------------------------------------------------------------

- (void) startAnimationTimer;
- (void) stopAnimationTimer;
- (void) fireAnimationTimer:(NSTimer *)aTimer;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- (NSImage *) animatedImage;
- (NSString *) description;

//----------------------------------------------------------------------
// State changes
//----------------------------------------------------------------------

- (void) enterInactiveState;
- (void) enterLaunching1State;
- (void) enterLaunching2State;
- (void) enterActiveState:(FIAppInfo *)anAppInfo;

//----------------------------------------------------------------------
// FITileView Delegate protocol
//----------------------------------------------------------------------

- (DockLocation) _adjustedLocationForOffset:(DockLocation)offset;
- (void) tileDidDragWindow:(NSWindow *)shadowWindow locationOffset:(DockLocation)offset;
- (void) tileDidDragToOffset:(DockLocation)offset;

@end
