#import "Fiend.h"

RCSID ("$Id: FITile.m,v 1.1.1.1 1998/09/03 01:14:51 nygard Exp $");

#import "FITile.h"

#import <AppKit/AppKit.h>
#import "FITileView.h"
#import "FIWindow.h"
#import "FIAppInfo.h"
#import "Fiendish.h"
#import "FiendWraps.h"
#import "FILevel.h"
#import "FIDock.h"
#import "FILauncher.h"

@interface NSWindow (Hidden)
- (void) _setAvoidsActivation:(char)flag;
@end

NSString *NSStringFromApplicationState (ApplicationState state)
{
    NSString *str;
    
    switch (state)
    {
      case ApplicationInactiveState:
          str = NSSTRINGIFY (ApplicationInactiveState);
          break;
          
      case ApplicationLaunching1State:
          str = NSSTRINGIFY (ApplicationLaunching1State);
          break;
          
      case ApplicationLaunching2State:
          str = NSSTRINGIFY (ApplicationLaunching2State);
          break;
          
      case ApplicationActiveState:
          str = NSSTRINGIFY (ApplicationActiveState);
          break;
          
      default:
          str = @"<unknown application state>";
    }

    return str;
}

//----------------------------------------------------------------------

@implementation FITile

#define PL_ApplicationPath @"ApplicationPath"
#define PL_ShouldAnimate @"ShouldAnimate"
#define PL_ShouldAutoLaunch @"ShouldAutoLaunch"
#define PL_XTileOrigin @"XTileOrigin"
#define PL_YTileOrigin @"YTileOrigin"

+ (FITile *) tileFromPropertyList:(NSDictionary *)propertyList
{
    FITile *newTile;
    id value;
    DockLocation aLocation;

    newTile = [[[FITile alloc] initWithRepresentedPath:[propertyList objectForKey:PL_ApplicationPath]] autorelease];

    value = [propertyList objectForKey:PL_ShouldAnimate];
    if (value != nil)
        [newTile setShouldAnimate:SNBoolFromString (value)];

    value = [propertyList objectForKey:PL_ShouldAutoLaunch];
    if (value != nil)
        [newTile setShouldAutoLaunch:SNBoolFromString (value)];

    aLocation.x = 0;
    aLocation.y = 0;

    value = [propertyList objectForKey:PL_XTileOrigin];
    if (value != nil)
        aLocation.x = [value intValue];
    
    value = [propertyList objectForKey:PL_YTileOrigin];
    if (value != nil)
        aLocation.y = [value intValue];

    [newTile setLocation:aLocation];

    [newTile createTileWindow];

    return newTile;
}

//----------------------------------------------------------------------

- (NSDictionary *) tilePropertyList
{
    NSMutableDictionary *propertyList;

    propertyList = [NSMutableDictionary dictionary];
    [propertyList setObject:path forKey:PL_ApplicationPath];

    [propertyList setObject:NSStringFromBool (shouldAnimate) forKey:PL_ShouldAnimate];
    [propertyList setObject:NSStringFromBool (shouldAutoLaunch) forKey:PL_ShouldAutoLaunch];

    // Reading the property list doesn't accepts an unquoted negative value, which is what gets
    // used by setting the object to the NSNumber.
    [propertyList setObject:[NSString stringWithFormat:@"%d", location.x] forKey:PL_XTileOrigin];
    [propertyList setObject:[NSString stringWithFormat:@"%d", location.y] forKey:PL_YTileOrigin];

    return propertyList;
}

//----------------------------------------------------------------------

- initWithRepresentedPath:(NSString *)aPath
{
    if ([super init] == nil)
        return nil;

    nonretainedLevel = nil;

    path = [aPath copy];
    shouldAutoLaunch = NO;
    location.x = 0;
    location.y = 0;

    window = nil;
    iconView = nil;

    shouldAutoLaunch = NO;
    shouldAnimate = NO;
    isPersistent = NO;

    isSelected = NO;

    state = ApplicationInactiveState;

    appInfoTimer = nil;
    appInfo = nil;

    animationTimer = nil;
    animationImage = nil;

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    [self stopAnimationTimer];

    [self stopAppInfoTimer];
    SNRelease (appInfo);

    SNRelease (path);
    SNRelease (window);

    [super dealloc];
}

//----------------------------------------------------------------------

- (NSString *) path
{
    return path;
}

//----------------------------------------------------------------------

- (NSString *) shortName
{
    return [[path lastPathComponent] stringByDeletingPathExtension];
}

//----------------------------------------------------------------------

- (NSWindow *) window
{
    return window;
}

//----------------------------------------------------------------------

- (FILevel *) level
{
    return nonretainedLevel;
}

//----------------------------------------------------------------------

- (void) setLevel:(FILevel *)newLevel
{
    nonretainedLevel = newLevel;
}

//----------------------------------------------------------------------

- (BOOL) shouldAutoLaunch
{
    return shouldAutoLaunch;
}

//----------------------------------------------------------------------

- (void) setShouldAutoLaunch:(BOOL)flag
{
    if (flag == shouldAutoLaunch)
        return;
    
    shouldAutoLaunch = flag;

    [iconView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (BOOL) shouldAnimate
{
    return shouldAnimate;
}
    
//----------------------------------------------------------------------

- (void) setShouldAnimate:(BOOL)flag
{
    if (flag == shouldAnimate)
        return;
    
    [iconView setNeedsDisplay:YES];

    shouldAnimate = flag;
    if (shouldAnimate == YES && state == ApplicationActiveState)
        [self startAnimationTimer];
    else if (shouldAnimate == NO)
        [self stopAnimationTimer];
}

//----------------------------------------------------------------------

- (BOOL) isPersistent
{
    return isPersistent;
}

//----------------------------------------------------------------------

- (void) setIsPersistent:(BOOL)flag
{
    if (flag == isPersistent)
        return;
    
    isPersistent = flag;

    [iconView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (BOOL) isRunning
{
    return (state == ApplicationLaunching1State) || (state == ApplicationLaunching2State) || (state == ApplicationActiveState);
}

//----------------------------------------------------------------------

- (BOOL) isSelected
{
    return isSelected;
}

//----------------------------------------------------------------------

- (void) setIsSelected:(BOOL)flag
{
    if (flag == isSelected)
        return;
    
    isSelected = flag;
    [iconView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (BOOL) isGhosted
{
    return (state == ApplicationLaunching1State) || (state == ApplicationLaunching2State);
}

//----------------------------------------------------------------------

- (DockLocation) location
{
    return location;
}

//----------------------------------------------------------------------

- (void) setLocation:(DockLocation)newLocation
{
    location = newLocation;
}

//----------------------------------------------------------------------

- (void) rotateLocationClockwise:(BOOL)clockwise
{
    DockLocation newLocation;

    if (clockwise == YES)
    {
        newLocation.x = location.y;
        newLocation.y = -location.x;
    }
    else
    {
        newLocation.x = -location.y;
        newLocation.y = location.x;
    }

    [self setLocation:newLocation];
}

//----------------------------------------------------------------------

- (void) setWindow:(NSWindow *)newWindow
{
    NSAssert (window == nil, @"expected window to be nil.");

    window = [newWindow retain];
}

//----------------------------------------------------------------------

- (void) createTileWindow
{
    NSRect frame;
    NSWorkspace *workspace;
    NSImage *image;

    frame = NSMakeRect (0, 0, STANDARD_TILE_WIDTH, STANDARD_TILE_HEIGHT);

    window = [[FIWindow alloc] initWithContentRect:frame
                               styleMask:NSBorderlessWindowMask
                               backing:NSBackingStoreBuffered
                               defer:NO];
    //[window setLevel:dockWindowLevel];
    [window _setAvoidsActivation:YES];

    workspace = [NSWorkspace sharedWorkspace];
    image = [workspace iconForFile:path];
    //NSLog (@"workspace image: %@", image);
    iconView = [[FITileView alloc] initWithFrame:NSMakeRect (0, 0, STANDARD_TILE_WIDTH, STANDARD_TILE_HEIGHT)];

    [iconView setImage:image];
    [window setContentView:iconView];

    [iconView setDelegate:self];
    [iconView setDatasource:self];
}

//----------------------------------------------------------------------

- (void) repositionRelativeToPoint:(NSPoint)origin
{
    origin.x += location.x * STANDARD_TILE_WIDTH;
    origin.y += location.y * STANDARD_TILE_HEIGHT;
    [window setFrameOrigin:origin];
}

//----------------------------------------------------------------------

- (void) hide
{
    [window orderOut:self];
}

//----------------------------------------------------------------------

- (void) unhide
{
    [window orderFront:self];
}

//----------------------------------------------------------------------

- (void) setTileWindowLevel:(int)level
{
    [window setLevel:level];
}

//----------------------------------------------------------------------

- (void) checkForActiveApplication:(NSDictionary *)activeApplications
{
    FIAppInfo *anAppInfo;
    
    anAppInfo = [activeApplications objectForKey:[self shortName]];
    if (anAppInfo != nil)
    {
        //NSLog (@"%@ is already active.", anAppInfo);
        [self enterActiveState:anAppInfo];
    }
}

//----------------------------------------------------------------------
// FITileView target methods
//----------------------------------------------------------------------

- (void) selectApplicationInWorkspace:sender
{
    BOOL flag;

    if ([[NSUserDefaults standardUserDefaults] boolForKey:DK_EnableSingleClick] == YES)
    {
        //flag = [[NSWorkspace sharedWorkspace] selectFile:path inFileViewerRootedAtPath:@""];
        flag = [[NSWorkspace sharedWorkspace] selectFile:path inFileViewerRootedAtPath:@"/"];
        //NSLog (@"flag: %@", (flag == YES) ? @"Yes" : @"No");
    }
}

//----------------------------------------------------------------------

- (void) autoLaunch
{
    if ([self shouldAutoLaunch] == YES && [self isRunning] == NO)
    {
        //NSLog (@"autolaunching %@", [self shortName]);
        [self enterLaunching1State];
        [iconView update];

        [[FILauncher sharedLauncher] launchApplication:path autolaunch:YES];
#if 0
        didLaunch = [[NSWorkspace sharedWorkspace] launchApplication:path showIcon:YES autolaunch:YES];
        if (didLaunch == NO)
        {
            NSLog (@"Autolaunch of %@ didn't take.", [self shortName]);
            [self enterInactiveState];
        }
#endif
    }
}

//----------------------------------------------------------------------

- (void) hideSourceTile
{
    if (appInfo != nil && [appInfo tokenWindowNumber] >= 0)
    {
        PSWMove (-1000, 1000, [appInfo tokenWindowNumber]);
    }
}

//----------------------------------------------------------------------

- (void) unhideSourceTile
{
    if (appInfo != nil && [appInfo tokenWindowNumber] >= 0)
    {
        PSWReplaceIcon2 ([appInfo tokenWindowNumber]);
    }
}

//----------------------------------------------------------------------

- (void) updateTileView
{
    [iconView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (void) launchApplication:sender hideOthers:(BOOL)hideOthers
{
    if ([self isRunning] == NO)
    {
        [self enterLaunching1State];
        [iconView update];

        [[FILauncher sharedLauncher] launchApplication:path autolaunch:NO];
#if 0
        didLaunch = [[NSWorkspace sharedWorkspace] launchApplication:path showIcon:YES autolaunch:NO];
        if (didLaunch == NO)
        {
            [self enterInactiveState];
        }
        else
        {
            // Activate the application.
            //NSLog (@"We want to activate, appInfo: %@", appInfo);
            //[NSApp hide:self];
        }
#endif
    }
    else
    {
        // Activate the application.
        if (appInfo != nil)
        {
            [appInfo activateApplicationAndHideOthers:hideOthers];
        }
        else
        {
            NSBeep ();
        }
    }
}

//----------------------------------------------------------------------

- (void) toggleTileSelected:sender
{
    [NSApp activateIgnoringOtherApps:YES];
    [NSApp sendAction:@selector (toggleSelection:) to:nil from:self];
    //[self setIsSelected:!isSelected];
}

//----------------------------------------------------------------------
// NSWorkspace notifications
//----------------------------------------------------------------------

- (void) applicationWillLaunch:(NSNotification *)aNotification
{
    NSString *launchedApplication;

    launchedApplication = [[aNotification userInfo] objectForKey:@"NSApplicationName"];
    if ([path isEqual:launchedApplication] == YES)
    {
        //NSLog (@"%@ will launch.", path);
        [self enterLaunching1State];
    }
}

//----------------------------------------------------------------------

- (void) applicationDidLaunch:(NSNotification *)aNotification
{
    NSString *launchedApplication;

    launchedApplication = [[aNotification userInfo] objectForKey:@"NSApplicationName"];
    if ([path isEqual:launchedApplication] == YES)
    {
        //NSLog (@"%@ did launch.", path);
        [self enterLaunching2State];
    }
}

//----------------------------------------------------------------------

- (void) applicationDidTerminate:(NSNotification *)aNotification
{
    NSString *launchedApplication;

    launchedApplication = [[aNotification userInfo] objectForKey:@"NSApplicationName"];
    if ([path isEqual:launchedApplication] == YES)
    {
        //NSLog (@"%@ did terminate.", path);
        [self enterInactiveState];
    }
}

//----------------------------------------------------------------------
// Fiend Launcher notifications
//----------------------------------------------------------------------

- (void) applicationLaunchFailed:(NSNotification *)aNotification
{
    NSString *applicationPath;

    applicationPath = [[aNotification userInfo] objectForKey:FILauncherApplicationPath];
    if ([path isEqual:applicationPath] == YES)
    {
        NSLog (@"Failed to launch: %@", applicationPath);
        [self enterInactiveState];
    }
}

//----------------------------------------------------------------------
// App Info timer
//----------------------------------------------------------------------

- (void) startAppInfoTimer
{
    // Replace any existing timer...
    [self stopAppInfoTimer];

    appInfoTimer = [[NSTimer scheduledTimerWithTimeInterval:0.2
                             target:self
                             selector:@selector (fireAppInfoTimer:)
                             userInfo:nil
                             repeats:YES] retain];
}

//----------------------------------------------------------------------

- (void) stopAppInfoTimer
{
    if (appInfoTimer != nil)
    {
        [appInfoTimer invalidate];
        SNRelease (appInfoTimer);
    }
}

//----------------------------------------------------------------------

- (void) fireAppInfoTimer:(NSTimer *)aTimer
{
    NSDictionary *activeApplications;
    FIAppInfo *anAppInfo;

    activeApplications = [FIAppInfo activeAppInfo];
    anAppInfo = [activeApplications objectForKey:[self shortName]];
    if (anAppInfo != nil)
    {
        //NSLog (@"Got it! %@", anAppInfo);
        [self enterActiveState:anAppInfo];
    }
    else
    {
        //NSLog (@"Couldn't get app info for '%@'.", [self shortName]);
    }
}

//----------------------------------------------------------------------
// Animation timer
//----------------------------------------------------------------------

- (void) startAnimationTimer
{
    // Replace any existing timer...
    [self stopAnimationTimer];

    animationTimer = [[NSTimer scheduledTimerWithTimeInterval:0.2
                               target:self
                               selector:@selector (fireAnimationTimer:)
                               userInfo:nil
                               repeats:YES] retain];
}

//----------------------------------------------------------------------

- (void) stopAnimationTimer
{
    if (animationTimer != nil)
    {
        [animationTimer invalidate];
        SNRelease (animationTimer);
    }

    SNRelease (animationImage);
}

//----------------------------------------------------------------------

- (void) fireAnimationTimer:(NSTimer *)aTimer
{
    int status;
    
    if (appInfo != nil)
    {
        if ([appInfo tokenWindowNumber] < 0)
        {
            // i.e. Fiend itself has window number of "-1"...
            [self stopAnimationTimer];
        }
        else
        {
            if (animationImage == nil)
            {
                animationImage = [[NSImage alloc] initWithSize:NSMakeSize (60, 60)];
            }

            [animationImage lockFocus];
            // Yes, this is (2,1), not (2, 2)...
            status = safeCopyWindow (2, 1, 60, 60, [appInfo tokenWindowNumber]);
            [animationImage unlockFocus];
            if (status != 0)
            {
                NSLog (@"safeCopyWindow() failed.");
            }
            else
            {
                [iconView setNeedsDisplay:YES];
            }
        }
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- (NSImage *) animatedImage
{
    return animationImage;
}

//----------------------------------------------------------------------

- (NSString *) description
{
    return [NSString stringWithFormat:@"[%d,%d]: %@, al:%d an:%d per:%d",
                     location.x, location.y, path, shouldAutoLaunch, shouldAnimate, isPersistent];
}

//----------------------------------------------------------------------
// State changes
//----------------------------------------------------------------------

- (void) enterInactiveState
{
    switch (state)
    {
      case ApplicationLaunching1State:
          NSAssert (appInfo == nil, @"Expected nil appInfo.");
          NSAssert (appInfoTimer == nil, @"Expected nil appInfoTimer.");
          break;
          
      case ApplicationLaunching2State:
          NSAssert (appInfo == nil, @"Expected nil appInfo.");
          [self stopAppInfoTimer];
          break;
          
      case ApplicationActiveState:
          NSAssert (appInfoTimer == nil, @"Expected nil appInfoTimer.");
          SNRelease (appInfo);
          [self stopAnimationTimer];
          break;
          
      case ApplicationInactiveState:
      default:
          NSAssert (appInfo == nil, @"Expected nil appInfo.");
          NSAssert (appInfoTimer == nil, @"Expected nil appInfoTimer.");
          break;
    }

    state = ApplicationInactiveState;
    [iconView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (void) enterLaunching1State
{
    NSAssert2 (state == ApplicationInactiveState || state == ApplicationLaunching1State,
               @"%@:[%@] Expected to be in ApplicationInactiveState or ApplicationLaunching1State.",
               [self shortName], NSStringFromApplicationState (state));
    NSAssert (appInfo == nil, @"Expected nil appInfo.");
    NSAssert (appInfoTimer == nil, @"Expected nil appInfoTimer.");

    state = ApplicationLaunching1State;
    [iconView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (void) enterLaunching2State
{
    // We don't appear to be getting the -applicationWillLaunch: notifications.
    NSAssert2 (state == ApplicationLaunching1State || state == ApplicationInactiveState,
               @"%@:[%@] Expected to be in either ApplicationInactiveState or ApplicationLaunching1State.",
               [self shortName], NSStringFromApplicationState (state));
    NSAssert (appInfo == nil, @"Expected nil appInfo.");
    NSAssert (appInfoTimer == nil, @"Expected nil appInfoTimer.");

    state = ApplicationLaunching2State;
    [self startAppInfoTimer];
    [iconView setNeedsDisplay:YES];
}

//----------------------------------------------------------------------

- (void) enterActiveState:(FIAppInfo *)anAppInfo
{
    NSAssert2 (state == ApplicationInactiveState || state == ApplicationLaunching2State,
               @"%@:[%@] Expected to be in either ApplicationInactiveState or ApplicationLaunching2State.",
               [self shortName], NSStringFromApplicationState (state));
    NSAssert (appInfo == nil, @"Expected nil appInfo.");

    if (state == ApplicationInactiveState)
    {
        NSAssert (appInfoTimer == nil, @"Expected nil appInfoTimer.");
    }

    [self stopAppInfoTimer];
    appInfo = [anAppInfo retain];

    [self hideSourceTile];

    state = ApplicationActiveState;
    [iconView setNeedsDisplay:YES];
    if ([self shouldAnimate] == YES)
        [self startAnimationTimer];

    if ([path isEqual:[[FILauncher sharedLauncher] lastLaunchPath]] == YES)
    {
        [appInfo activateApplicationAndHideOthers:NO];
        [[FILauncher sharedLauncher] setLastLaunchPath:nil];
    }
    
}

//----------------------------------------------------------------------
// FITileView Delegate protocol
//----------------------------------------------------------------------

- (DockLocation) _adjustedLocationForOffset:(DockLocation)offset
{
    NSPoint windowOrigin;
    DockLocation adjustedLocation;
    FIDock *dock;
    NSPoint dockWindowOrigin;
    
    dock = [nonretainedLevel dock];
    adjustedLocation.x = location.x + offset.x;
    adjustedLocation.y = location.y + offset.y;

    windowOrigin = [window frame].origin;
    dockWindowOrigin.x = windowOrigin.x - location.x * 64;
    dockWindowOrigin.y = windowOrigin.y - location.y * 64;

    if ((offset.x != 0 || offset.y != 0) && [dock isLocationAvailable:adjustedLocation] == NO)
    {
        //NSLog (@"(%d,%d) is not available...", adjustedLocation.x, adjustedLocation.y);
        adjustedLocation = [dock nextAvailableDockLocation:dockWindowOrigin orLocation:location];
        //NSLog (@"but (%d,%d) is.", adjustedLocation.x, adjustedLocation.y);
    }

    return adjustedLocation;
}

//----------------------------------------------------------------------

- (void) tileDidDragWindow:(NSWindow *)shadowWindow locationOffset:(DockLocation)offset
{
    NSPoint shadowWindowOrigin;
    DockLocation adjustedLocation;
    
    shadowWindowOrigin = [window frame].origin;
    adjustedLocation = [self _adjustedLocationForOffset:offset];

    // If the adjusted location is (0,0), show a skull&crossbones above the tile at its original location.

    //shadowWindowOrigin.x = (shadowWindowOrigin.x - location.x * 64) + adjustedLocation.x * 64;
    //shadowWindowOrigin.y = (shadowWindowOrigin.y - location.y * 64) + adjustedLocation.y * 64;
    shadowWindowOrigin.x += (adjustedLocation.x - location.x) * 64;
    shadowWindowOrigin.y += (adjustedLocation.y - location.y) * 64;

    [shadowWindow setFrameOrigin:shadowWindowOrigin];

    // If newLocation is available, move window there
    // If it is unavailable, find next available location (count original location as free), and move it there.
}

//----------------------------------------------------------------------

- (void) tileDidDragToOffset:(DockLocation)offset
{
    // If the adjusted location is (0,0), remove the tile from the level.

    location = [self _adjustedLocationForOffset:offset];

    // Could now change this to [dock updateChangeCount] and remove from FILevel...
    [nonretainedLevel updateChangeCount];
}

@end
