#import "Fiend.h"

RCSID ("$Id: FITileView.m,v 1.1.1.1 1998/09/03 01:14:52 nygard Exp $");

#import "FITileView.h"
#import "FiendWraps.h"

static NSImage *_blankTileImage = nil;
static NSImage *_dotsImage = nil;

@implementation FITileView

+ (void) initialize
{
    NSBundle *thisBundle;

    if (self == [FITileView class])
    {
        thisBundle = [NSBundle bundleForClass:self];
        NSAssert (thisBundle != nil, @"Could not get bundle.");

        _blankTileImage = [[NSImage imageNamed:@"BlankTile.tiff" fromBundle:thisBundle] retain];
        _dotsImage = [[NSImage imageNamed:@"dots.tiff" fromBundle:thisBundle] retain];
    }
}

//----------------------------------------------------------------------

- initWithFrame:(NSRect)frameRect
{
    if ([super initWithFrame:frameRect] == nil)
        return nil;

    image = nil;
    compositePoint = NSZeroPoint;

    nonretainedDelegate = nil;

    return self;
}

//----------------------------------------------------------------------

- (void) dealloc
{
    SNRelease (image);

    [super dealloc];
}

//----------------------------------------------------------------------

- (NSImage *) image
{
    return image;
}

//----------------------------------------------------------------------

- (void) setImage:(NSImage *)newImage
{
    NSSize imageSize;
    
    if (newImage == image)
        return;

    SNRelease (image);
    image = [newImage copy];

    if (image != nil)
    {
        imageSize = [newImage size];

        // Don't scale smaller images, but be sure to shrink larger ones.
        // XXX: Should we worry about proportionality?
        if (imageSize.height > 48 || imageSize.width > 48)
        {
            [image setScalesWhenResized:YES];
            [image setSize:NSMakeSize (48, 48)];
            compositePoint = NSMakePoint (8, 8);
        }
        else
        {
            compositePoint = NSMakePoint ((64 - imageSize.width) / 2, (64 - imageSize.height) / 2);
        }
    }
}

//----------------------------------------------------------------------

- delegate
{
    return nonretainedDelegate;
}

//----------------------------------------------------------------------

- (void) setDelegate:newDelegate
{
    if (newDelegate == nonretainedDelegate)
        return;

    nonretainedDelegate = newDelegate;
}

//----------------------------------------------------------------------

- datasource
{
    return nonretainedDatasource;
}

//----------------------------------------------------------------------

- (void) setDatasource:newDatasource
{
    if (newDatasource == nonretainedDatasource)
        return;

    nonretainedDatasource = newDatasource;
}

//----------------------------------------------------------------------

void SNHighlightRect (NSRect rect)
{
    rect = NSInsetRect (rect, 2, 2);
    PSgsave ();
    PSsetlinewidth (4);
    [[NSColor blackColor] set];
    PSrectstroke (rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
    PSsetlinewidth (2);
    [[NSColor darkGrayColor] set];
    PSrectstroke (rect.origin.x, rect.origin.y, rect.size.width, rect.size.height);
    PSgrestore ();
}

//----------------------------------------------------------------------

- (void) drawRect:(NSRect)rect
{
    NSUserDefaults *defaults;
    NSImage *animatedImage;

    defaults = [NSUserDefaults standardUserDefaults];
    animatedImage = [self animatedImage];

    if (_blankTileImage != nil)
    {
        [_blankTileImage compositeToPoint:NSZeroPoint operation:NSCompositeSourceOver];
    }

    if (/*[self isRunning] == NO || */animatedImage == nil)
    {
        if (image != nil)
        {
            //[image compositeToPoint:NSMakePoint (8, 8) operation:NSCompositeSourceOver];
            [image compositeToPoint:compositePoint operation:NSCompositeSourceOver];
        }

        if (_dotsImage != nil && [self isRunning] == NO)
        {
            [_dotsImage compositeToPoint:NSZeroPoint operation:NSCompositeSourceOver];
        }

        if ([self isGhosted] == YES)
        {
            [[NSColor whiteColor] set];
            //PSsetalpha (0.666);
            PSsetalpha (0.5);
            PScompositerect (2, 2, 60, 60, NSCompositeSourceOver);
        }
    }
    else
    {
        //[animatedImage compositeToPoint:NSMakePoint (0, 0) operation:NSCompositeSourceOver];
        [animatedImage compositeToPoint:NSMakePoint (2, 2) operation:NSCompositeCopy];
    }

    if ([defaults boolForKey:DK_ShowFlagCues] == YES)
    {
        PSWMarkStatus (0, 64, 64, YES, [self shouldAutoLaunch], [self shouldAnimate], [self isPersistent]);
    }

    if ([self isSelected] == YES)
    {
        SNHighlightRect ([self bounds]);
    }
}

//----------------------------------------------------------------------

- (void) mouseDown:(NSEvent *)theEvent
{
    NSPoint originalWindowOrigin, originalScreenPoint;
    NSWindow *window;
    NSWindow *dragWindow, *shadowWindow;

    NSPoint mouseLocation;
    NSPoint newWindowOrigin;

    unsigned int modifierFlags;
    int clickCount;

    BOOL stopTracking;
    NSPoint mouseDownPoint;
    NSRect currentRect;
    //NSPoint shadowWindowOrigin;

    modifierFlags = [theEvent modifierFlags];
    clickCount = [theEvent clickCount];

    if (clickCount == 1)
    {
        window = [self window];
        originalWindowOrigin = [window frame].origin;
        originalScreenPoint = [window convertBaseToScreen:[theEvent locationInWindow]];

        //SNLogPoint (originalWindowOrigin);
        //SNLogPoint (originalScreenPoint);

        // XXX: Might we just be able to peek at the next event?
        theEvent = [window nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
        if ([theEvent type] == NSLeftMouseDragged)
        {
            DockLocation locationOffset;

            //----------
            shadowWindow = [self _tileWindow:YES];
            [shadowWindow orderFront:self];
            dragWindow = [self _tileWindow:NO];
            [dragWindow orderFront:self];
            //----------
            [window orderOut:self];

            //theEvent = [window nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
            mouseDownPoint = [window convertBaseToScreen:[theEvent locationInWindow]];
            mouseDownPoint.x -= 32;
            mouseDownPoint.y -= 32;
            currentRect.origin = mouseDownPoint;
            currentRect.size = NSMakeSize (64, 64);

            stopTracking = NO;
            while (stopTracking == NO)
            {
                mouseLocation = [window convertBaseToScreen:[theEvent locationInWindow]];
                //SNLogPoint (mouseLocation);
                locationOffset.x = floor ((mouseLocation.x - mouseDownPoint.x) / 64);
                locationOffset.y = floor ((mouseLocation.y - mouseDownPoint.y) / 64);
                currentRect.origin.x = mouseDownPoint.x + locationOffset.x * 64;
                currentRect.origin.y = mouseDownPoint.y + locationOffset.y * 64;
                //NSLog (@"offset: (%d,%d)", locationOffset.x, locationOffset.y);
                [nonretainedDelegate tileDidDragWindow:shadowWindow locationOffset:locationOffset];

                //----------
                while ([theEvent type] == NSLeftMouseDragged)
                {
                    mouseLocation = [window convertBaseToScreen:[theEvent locationInWindow]];
                    //SNLogPoint (mouseLocation);
                    if (NSPointInRect (mouseLocation, currentRect) == NO)
                        break;

                    newWindowOrigin.x = originalWindowOrigin.x + mouseLocation.x - originalScreenPoint.x;
                    newWindowOrigin.y = originalWindowOrigin.y + mouseLocation.y - originalScreenPoint.y;

                    [dragWindow setFrameOrigin:newWindowOrigin];

                    theEvent = [window nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
                }

                if ([theEvent type] == NSLeftMouseUp)
                    stopTracking = YES;
                else
                    theEvent = [window nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
                //----------
            }

#if 0
                //stopTracking = [self trackMouse:theEvent inRect:currentRect];
                stopTracking = [self dragWindow:dragWindow
                                     fromOrigin:originalWindowOrigin
                                     inRect:currentRect
                                     firstEvent:theEvent];
#endif

            [window setFrameOrigin:[shadowWindow frame].origin];
            [nonretainedDelegate tileDidDragToOffset:locationOffset];
            [window orderFront:self];
            [dragWindow orderOut:self];
            [shadowWindow orderOut:self];
        }
        else
        {
            if ([NSApp nextEventMatchingMask:NSLeftMouseDownMask
                       untilDate:[NSDate dateWithTimeIntervalSinceNow:0.2]
                       inMode:NSModalPanelRunLoopMode
                       dequeue:NO] == nil)
            {
                // It looks like this will be a lone single click.
                [self singleClicked];
            }
            else
            {
                // Double click.
                // Let the code below handle it.
            }
            // If we actually dragged, gobble up the mouse up event, otherwise pass it on.
            [self mouseUp:theEvent];
        }
    }
    else if (clickCount == 2)
    {
        BOOL hideOthers;

        hideOthers = (modifierFlags & NSCommandKeyMask) ? YES : NO;
        [self launchApplicationAndHideOthers:hideOthers];
    }
    else
    {
        [super mouseDown:theEvent];
    }
}

//----------------------------------------------------------------------

#if 0
//- (BOOL) trackMouse:(NSEvent *)theEvent inRect:(NSRect)aRect
- (BOOL) dragWindow:(NSWindow *)dragWindow
         fromOrigin:(NSPoint)originalWindowOrigin
             inRect:(NSRect)aRect
         firstEvent:(NSEvent *)theEvent
{
    NSWindow *window;
    BOOL gotMouseUp;
    NSPoint mouseScreenLocation;
    NSPoint newWindowOrigin;

    window = [self window];
    gotMouseUp = NO;

    //SNLogRect (aRect);

    while ([theEvent type] == NSLeftMouseDragged)
    {
        mouseScreenLocation = [window convertBaseToScreen:[theEvent locationInWindow]];
        //SNLogPoint (mouseScreenLocation);
        if (NSPointInRect (mouseScreenLocation, aRect) == NO)
            break;

        newWindowOrigin.x = originalWindowOrigin.x + mouseScreenLocation.x - originalScreenPoint.x;
        newWindowOrigin.y = originalWindowOrigin.y + mouseScreenLocation.y - originalScreenPoint.y;

        [dragWindow setFrameOrigin:newWindowOrigin];

        theEvent = [window nextEventMatchingMask:(NSLeftMouseUpMask | NSLeftMouseDraggedMask)];
    }

    if ([theEvent type] == NSLeftMouseUp)
        gotMouseUp = YES;

    return gotMouseUp;
}
#endif
//----------------------------------------------------------------------

- (void) mouseUp:(NSEvent *)theEvent
{
    unsigned int modifierFlags;

    modifierFlags = [theEvent modifierFlags];

    if (modifierFlags & NSShiftKeyMask)
    {
        [self toggleTileSelected];
    }
    else
    {
        [NSApp sendAction:@selector (deselectAllTiles:) to:nil from:self];
        //[super mouseUp:theEvent];
    }
}

//----------------------------------------------------------------------

- (void) update
{
    [self displayIfNeeded];
    [self lockFocus];
    PSWait ();
    [self unlockFocus];
}

//----------------------------------------------------------------------
// Covers for datasource methods
//----------------------------------------------------------------------

- (BOOL) shouldAutoLaunch
{
    BOOL flag;

    flag = NO;
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (shouldAutoLaunch)] == YES)
    {
        flag = [nonretainedDatasource shouldAutoLaunch];
    }

    return flag;
}

//----------------------------------------------------------------------

- (BOOL) shouldAnimate
{
    BOOL flag;

    flag = NO;
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (shouldAnimate)] == YES)
    {
        flag = [nonretainedDatasource shouldAnimate];
    }

    return flag;
}

//----------------------------------------------------------------------

- (BOOL) isPersistent
{
    BOOL flag;

    flag = NO;
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (isPersistent)] == YES)
    {
        flag = [nonretainedDatasource isPersistent];
    }

    return flag;
}

//----------------------------------------------------------------------

- (BOOL) isRunning
{
    BOOL flag;

    flag = NO;
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (isRunning)] == YES)
    {
        flag = [nonretainedDatasource isRunning];
    }

    return flag;
}

//----------------------------------------------------------------------

- (BOOL) isSelected
{
    BOOL flag;

    flag = NO;
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (isSelected)] == YES)
    {
        flag = [nonretainedDatasource isSelected];
    }

    return flag;
}

//----------------------------------------------------------------------

- (BOOL) isGhosted
{
    BOOL flag;

    flag = NO;
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (isGhosted)] == YES)
    {
        flag = [nonretainedDatasource isGhosted];
    }

    return flag;
}

//----------------------------------------------------------------------

- (NSImage *) animatedImage
{
    NSImage *animatedImage;

    animatedImage = nil;
    if (nonretainedDatasource != nil && [nonretainedDatasource respondsToSelector:@selector (animatedImage)] == YES)
    {
        animatedImage = [nonretainedDatasource animatedImage];
    }

    return animatedImage;
}

//----------------------------------------------------------------------
// Covers for delegate messages
//----------------------------------------------------------------------

- (void) launchApplicationAndHideOthers:(BOOL)hideOthers
{
    if (nonretainedDelegate != nil && [nonretainedDelegate respondsToSelector:@selector (launchApplication:hideOthers:)] == YES)
    {
        [nonretainedDelegate launchApplication:self hideOthers:hideOthers];
    }
}

//----------------------------------------------------------------------

- (void) singleClicked
{
    if (nonretainedDelegate != nil && [nonretainedDelegate respondsToSelector:@selector (selectApplicationInWorkspace:)] == YES)
    {
        [nonretainedDelegate selectApplicationInWorkspace:self];
    }
}

//----------------------------------------------------------------------

- (void) toggleTileSelected
{
    if (nonretainedDelegate != nil && [nonretainedDelegate respondsToSelector:@selector (toggleTileSelected:)] == YES)
    {
        [nonretainedDelegate toggleTileSelected:self];
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- (NSWindow *) _tileWindow:(BOOL)isShadowedWindow
{
    NSWindow *dockedWindow, *tileWindow;
    NSRect windowFrame, imageViewFrame;
    NSImageView *imageView;
    NSImage *windowImage;

    dockedWindow = [self window];
    windowFrame = [dockedWindow frame];

    tileWindow = [[[NSWindow alloc] initWithContentRect:windowFrame
                                    styleMask:NSBorderlessWindowMask
                                    backing:NSBackingStoreBuffered
                                    defer:NO] autorelease];
    if (isShadowedWindow == YES)
        [tileWindow setLevel:[dockedWindow level]];
    else
        [tileWindow setLevel:NSDockWindowLevel + 1];

    imageViewFrame.origin = NSZeroPoint;
    imageViewFrame.size = windowFrame.size;
        
    windowImage = [[NSImage alloc] initWithSize:imageViewFrame.size];
    [windowImage lockFocus];
    [self drawRect:imageViewFrame];
    if (isShadowedWindow == YES)
    {
        [[NSColor whiteColor] set];
        //[[NSColor greenColor] set];
        PSsetalpha (0.666);
        PScompositerect (2, 2, 60, 60, NSCompositeSourceOver);
    }
    [windowImage unlockFocus];

    imageView = [[NSImageView alloc] initWithFrame:imageViewFrame];
    [imageView setImageAlignment:NSImageAlignBottomLeft];
    [imageView setImageFrameStyle:NSImageFrameNone];
    [imageView setImageScaling:NSScaleNone];
    [imageView setImage:windowImage];
    [windowImage release];

    [tileWindow setContentView:imageView];
    [imageView release];

    return tileWindow;
}

@end
